"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Iterable, List, Literal, Optional, Union, cast
from unittest import IsolatedAsyncioTestCase, mock

import httpx
import openai
from botbuilder.core import TurnContext
from dataclasses_json import DataClassJsonMixin, dataclass_json
from openai._base_client import AsyncPaginator
from openai.pagination import AsyncCursorPage
from openai.types import beta
from openai.types.beta import AssistantTool, Thread
from openai.types.beta.threads import (
    Message,
    RequiredActionFunctionToolCall,
    Run,
    Text,
    TextContentBlock,
    required_action_function_tool_call,
    run,
    run_submit_tool_outputs_params,
)

from teams.ai.actions.action_types import ActionTypes
from teams.ai.planners.assistants_planner import (
    AssistantsPlanner,
    OpenAIAssistantsOptions,
)
from teams.ai.planners.plan import PredictedDoCommand, PredictedSayCommand
from teams.app_error import ApplicationError
from teams.state import TurnState
from teams.state.conversation_state import ConversationState
from teams.state.temp_state import TempState
from teams.state.user_state import UserState

ASSISTANT_ID = "assistant_id"
ASSISTANT_MODEL = "test model"
ASSISTANT_INSTRUCTIONS = "instructions"
ASSISTANT_TOOLS: List[AssistantTool] = []

RunStatuses = Literal[
    "queued",
    "in_progress",
    "requires_action",
    "cancelling",
    "cancelled",
    "failed",
    "completed",
    "expired",
]


@dataclass_json
@dataclass
class MockAsyncPaginator(DataClassJsonMixin):
    data: Union[List[Run], List[Message]]
    first_id: str
    last_id: str
    has_more: bool


class MockAsyncRuns:
    _threads: List[Thread]
    _messages: Dict[str, List[Message]]
    _runs: Dict[str, List[Run]]
    remaining_actions: List[run.RequiredAction]
    remaining_run_status: List[RunStatuses]
    remaining_messages: List[str]

    def __init__(
        self,
        shared_threads: List[Thread],
        shared_messages: Dict[str, List[Message]],
        shared_runs: Dict[str, List[Run]],
        remaining_actions: List[run.RequiredAction],
        remaining_run_status: List[RunStatuses],
        remaining_messages: List[str],
    ) -> None:
        self._threads = shared_threads
        self._messages = shared_messages
        self._runs = shared_runs
        self.remaining_actions = remaining_actions
        self.remaining_run_status = remaining_run_status
        self.remaining_messages = remaining_messages

    async def retrieve(
        self,
        run_id: str,
        *,
        thread_id: str,
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> Run:
        # pylint: disable=unused-argument

        if len(self._runs[thread_id]) == 0:
            raise ValueError("no more runs left")

        run_status = self.remaining_run_status.pop(0)
        index = -1

        for runs in self._runs[thread_id]:
            index += 1
            if runs.id == run_id:
                break
            if index == (len(self._runs[thread_id]) - 1):
                index = -1
                break

        curr_run = self._runs[thread_id][index]
        curr_run.status = run_status
        return curr_run

    async def submit_tool_outputs(
        self,
        run_id: str,
        *,
        thread_id: str,
        tool_outputs: Iterable[run_submit_tool_outputs_params.ToolOutput],
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> Run:
        # pylint: disable=unused-argument
        return await self.retrieve(thread_id=thread_id, run_id=run_id)

    async def create(
        self,
        thread_id: str,
        *,
        assistant_id: str,
        additional_instructions: Union[
            Optional[str], openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
        instructions: Union[Optional[str], openai._types.NotGiven] = openai._types.NOT_GIVEN,
        metadata: Union[Optional[object], openai._types.NotGiven] = openai._types.NOT_GIVEN,
        model: Union[Optional[str], openai._types.NotGiven] = openai._types.NOT_GIVEN,
        tools: Union[
            Optional[Iterable[AssistantTool]], openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> Run:
        # pylint: disable=unused-argument
        remaining_actions: Optional[run.RequiredAction] = None

        if len(self.remaining_actions) > 0:
            remaining_actions = self.remaining_actions.pop(0)

        new_run = Run(
            id=str(datetime.now()),
            thread_id=thread_id,
            assistant_id=assistant_id,
            status="in_progress",
            required_action=remaining_actions if remaining_actions else None,
            model=ASSISTANT_MODEL if ASSISTANT_MODEL else "test-model",
            instructions=ASSISTANT_INSTRUCTIONS if ASSISTANT_INSTRUCTIONS else "instructions",
            tools=ASSISTANT_TOOLS if ASSISTANT_TOOLS else [],
            created_at=235312,
            expires_at=2317312,
            started_at=2312812,
            completed_at=61312321,
            cancelled_at=5312321,
            failed_at=31712312,
            object="thread.run",
        )

        if thread_id in self._runs:
            self._runs[thread_id].append(new_run)
        else:
            self._runs[thread_id] = [new_run]

        return new_run

    async def list(
        self,
        thread_id: str,
        *,
        after: Union[str, openai._types.NotGiven] = openai._types.NOT_GIVEN,
        before: Union[str, openai._types.NotGiven] = openai._types.NOT_GIVEN,
        limit: Union[int, openai._types.NotGiven],
        order: Union[Literal["asc", "desc"], openai._types.NotGiven] = openai._types.NOT_GIVEN,
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> AsyncPaginator[Run, AsyncCursorPage[Run]]:
        # pylint: disable=unused-argument
        runs = self._runs[thread_id] if thread_id in self._runs else []

        result = MockAsyncPaginator(data=runs, first_id="", last_id="", has_more=False)
        return cast(AsyncPaginator, result)


class MockAsyncMessages:
    _threads: List[Thread]
    _messages: Dict[str, List[Message]]
    _runs: Dict[str, List[Run]]
    remaining_actions: List[run.RequiredAction]
    remaining_run_status: List[RunStatuses]
    remaining_messages: List[str]

    def __init__(
        self,
        shared_threads: List[Thread],
        shared_messages: Dict[str, List[Message]],
        shared_runs: Dict[str, List[Run]],
        remaining_actions: List[run.RequiredAction],
        remaining_run_status: List[RunStatuses],
        remaining_messages: List[str],
    ) -> None:
        self._threads = shared_threads
        self._messages = shared_messages
        self._runs = shared_runs
        self.remaining_actions = remaining_actions
        self.remaining_run_status = remaining_run_status
        self.remaining_messages = remaining_messages

    async def create(
        self,
        thread_id: str,
        *,
        content: str,
        role: Literal["user"],
        file_ids: Union[List[str], openai._types.NotGiven] = openai._types.NOT_GIVEN,
        metadata: Union[Optional[object], openai._types.NotGiven] = openai._types.NOT_GIVEN,
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> Message:
        # pylint: disable=unused-argument
        new_message = Message(
            id=str(datetime.now()),
            status="completed",
            created_at=123,
            thread_id=thread_id,
            role=role,
            content=[TextContentBlock(type="text", text=Text(annotations=[], value=content))],
            assistant_id=ASSISTANT_ID,
            object="thread.message",
        )

        if thread_id in self._messages:
            self._messages[thread_id].append(new_message)
        else:
            self._messages[thread_id] = [new_message]
        return new_message

    async def list(
        self,
        thread_id: str,
        *,
        after: Union[str, openai._types.NotGiven] = openai._types.NOT_GIVEN,
        before: Union[str, openai._types.NotGiven] = openai._types.NOT_GIVEN,
        limit: Union[int, openai._types.NotGiven] = openai._types.NOT_GIVEN,
        order: Union[Literal["asc", "desc"], openai._types.NotGiven] = openai._types.NOT_GIVEN,
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> AsyncPaginator[Message, AsyncCursorPage[Message]]:
        # pylint: disable=unused-argument
        while len(self.remaining_messages) > 0:
            next_message = self.remaining_messages.pop(0)
            await self.create(
                thread_id=thread_id, role="user", content=next_message, file_ids=[], metadata=None
            )

        last_message_id = before
        index = -1

        for threads in self._messages[thread_id]:
            index += 1
            if threads.id == last_message_id:
                break
            if index == (len(self._messages[thread_id]) - 1):
                index = -1
                break

        filtered_messages = self._messages[thread_id][index + 1 :]

        filtered_messages.reverse()

        result = MockAsyncPaginator(data=filtered_messages, first_id="", last_id="", has_more=False)

        return cast(AsyncPaginator, result)


class MockAsyncAssistants:
    _threads: List[Thread]
    _messages: Dict[str, List[Message]]
    _runs: Dict[str, List[Run]]
    remaining_actions: List[run.RequiredAction]
    remaining_run_status: List[RunStatuses]
    remaining_messages: List[str]

    def __init__(
        self,
        shared_threads: List[Thread],
        shared_messages: Dict[str, List[Message]],
        shared_runs: Dict[str, List[Run]],
        remaining_actions: List[run.RequiredAction],
        remaining_run_status: List[RunStatuses],
        remaining_messages: List[str],
    ) -> None:
        self._threads = shared_threads
        self._messages = shared_messages
        self._runs = shared_runs
        self.remaining_actions = remaining_actions
        self.remaining_run_status = remaining_run_status
        self.remaining_messages = remaining_messages

    async def create(
        self,
        *,
        model: str,
        description: Union[Optional[str], openai._types.NotGiven] = "",
        file_ids: Union[List[str], openai._types.NotGiven] = [],
        instructions: Union[Optional[str], openai._types.NotGiven] = None,
        metadata: Union[Optional[object], openai._types.NotGiven] = openai._types.NotGiven,
        name: Union[Optional[str], openai._types.NotGiven] = "",
        tools: Union[Iterable[AssistantTool], openai._types.NotGiven] = openai._types.NOT_GIVEN,
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> beta.Assistant:
        # pylint: disable=unused-argument
        # pylint: disable=dangerous-default-value
        return beta.Assistant(
            id=ASSISTANT_ID,
            created_at=3232,
            instructions=ASSISTANT_INSTRUCTIONS,
            model=ASSISTANT_MODEL,
            object="assistant",
            tools=[],
        )


class MockAsyncThreads:
    runs: MockAsyncRuns
    messages: MockAsyncMessages
    _threads: List[Thread]
    _messages: Dict[str, List[Message]]
    _runs: Dict[str, List[Run]]
    remaining_actions: List[run.RequiredAction]
    remaining_run_status: List[RunStatuses]
    remaining_messages: List[str]

    def __init__(
        self,
        shared_threads: List[Thread],
        shared_messages: Dict[str, List[Message]],
        shared_runs: Dict[str, List[Run]],
        remaining_actions: List[run.RequiredAction],
        remaining_run_status: List[RunStatuses],
        remaining_messages: List[str],
    ) -> None:
        self._threads = shared_threads
        self._messages = shared_messages
        self._runs = shared_runs
        self.remaining_actions = remaining_actions
        self.remaining_run_status = remaining_run_status
        self.remaining_messages = remaining_messages

        self.runs = MockAsyncRuns(
            shared_threads=self._threads,
            shared_messages=self._messages,
            shared_runs=self._runs,
            remaining_actions=self.remaining_actions,
            remaining_run_status=self.remaining_run_status,
            remaining_messages=self.remaining_messages,
        )
        self.messages = MockAsyncMessages(
            shared_threads=self._threads,
            shared_messages=self._messages,
            shared_runs=self._runs,
            remaining_actions=self.remaining_actions,
            remaining_run_status=self.remaining_run_status,
            remaining_messages=self.remaining_messages,
        )

    async def create(
        self,
        *,
        messages: Union[
            Iterable[beta.thread_create_params.Message], openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
        metadata: Union[Optional[object], openai._types.NotGiven] = openai._types.NOT_GIVEN,
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> Thread:
        # pylint: disable=unused-argument
        new_thread = Thread(
            id=str(datetime.now()),
            created_at=2312312,
            metadata=metadata if metadata else None,
            object="thread",
        )
        new_messages: List[Message] = []

        if isinstance(messages, Iterable):
            for message in messages:
                new_message = Message(
                    role=message["role"],
                    id=str(datetime.now()),
                    status="completed",
                    created_at=23123,
                    thread_id=new_thread.id,
                    assistant_id=ASSISTANT_ID,
                    content=[
                        TextContentBlock(
                            type="text",
                            text=Text(annotations=[], value=cast(str, message["content"])),
                        )
                    ],
                    object="thread.message",
                )
                new_messages.append(new_message)

        self._messages[new_thread.id] = new_messages
        self._threads.append(new_thread)
        return new_thread


class MockAsyncBeta:
    _threads: List[Thread]
    _messages: Dict[str, List[Message]]
    _runs: Dict[str, List[Run]]
    remaining_actions: List[run.RequiredAction]
    remaining_run_status: List[RunStatuses]
    remaining_messages: List[str]
    assistants: MockAsyncAssistants
    threads: MockAsyncThreads

    def __init__(
        self,
        shared_threads: List[Thread],
        shared_messages: Dict[str, List[Message]],
        shared_runs: Dict[str, List[Run]],
        remaining_actions: List[run.RequiredAction],
        remaining_run_status: List[RunStatuses],
        remaining_messages: List[str],
    ) -> None:
        self._threads = shared_threads
        self._messages = shared_messages
        self._runs = shared_runs
        self.remaining_actions = remaining_actions
        self.remaining_run_status = remaining_run_status
        self.remaining_messages = remaining_messages
        self.assistants = MockAsyncAssistants(
            shared_threads=self._threads,
            shared_messages=self._messages,
            shared_runs=self._runs,
            remaining_actions=self.remaining_actions,
            remaining_run_status=self.remaining_run_status,
            remaining_messages=self.remaining_messages,
        )
        self.threads = MockAsyncThreads(
            shared_threads=self._threads,
            shared_messages=self._messages,
            shared_runs=self._runs,
            remaining_actions=self.remaining_actions,
            remaining_run_status=self.remaining_run_status,
            remaining_messages=self.remaining_messages,
        )


class MockAsyncOpenAI:
    beta: MockAsyncBeta
    _threads: List[Thread]
    _messages: Dict[str, List[Message]]
    _runs: Dict[str, List[Run]]
    remaining_actions: List[run.RequiredAction]
    remaining_run_status: List[RunStatuses]
    remaining_messages: List[str]

    def __init__(
        self,
        shared_threads: List[Thread] = [],
        shared_messages: Dict[str, List[Message]] = {},
        shared_runs: Dict[str, List[Run]] = {},
        remaining_actions: List[run.RequiredAction] = [],
        remaining_run_status: List[RunStatuses] = [],
        remaining_messages: List[str] = [],
    ) -> None:
        # pylint: disable=dangerous-default-value
        self._threads = shared_threads
        self._messages = shared_messages
        self._runs = shared_runs
        self.remaining_actions = remaining_actions
        self.remaining_run_status = remaining_run_status
        self.remaining_messages = remaining_messages

        self.beta = MockAsyncBeta(
            self._threads,
            self._messages,
            self._runs,
            remaining_actions,
            remaining_run_status,
            remaining_messages,
        )


class TestAssistantsPlanner(IsolatedAsyncioTestCase):
    def create_mock_context(
        self, channel_id="channel1", bot_id="bot1", conversation_id="conversation1", user_id="user1"
    ):
        context = mock.MagicMock()
        context.activity.channel_id = channel_id
        context.activity.recipient.id = bot_id
        context.activity.conversation.id = conversation_id
        context.activity.from_property.id = user_id
        return context

    async def create_state(
        self, context: TurnContext
    ) -> TurnState[ConversationState, UserState, TempState]:
        state = await TurnState[ConversationState, UserState, TempState].load(context)
        state.temp.input = ""
        state.temp.input_files = []
        state.temp.last_output = ""
        state.temp.action_outputs = {}
        state.temp.auth_tokens = {}
        return state

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAI())
    async def test_create_openai_assistant(self, mock_async_openai):
        params = beta.AssistantCreateParams(model="123")

        assistant = await AssistantsPlanner.create_assistant(
            api_key="", api_version="", organization="", endpoint="", request=params
        )

        self.assertTrue(mock_async_openai.called)
        self.assertEqual(assistant.id, ASSISTANT_ID)
        self.assertEqual(assistant.model, ASSISTANT_MODEL)

    @mock.patch("openai.AsyncAzureOpenAI", return_value=MockAsyncOpenAI())
    async def test_create_azure_openai_assistant(self, mock_async_azure_openai):
        params = beta.AssistantCreateParams(model="123")

        assistant = await AssistantsPlanner.create_assistant(
            api_key="",
            api_version="",
            organization="",
            endpoint="this is my endpoint",
            request=params,
        )

        self.assertTrue(mock_async_azure_openai.called)
        self.assertEqual(assistant.id, ASSISTANT_ID)
        self.assertEqual(assistant.model, ASSISTANT_MODEL)

    @mock.patch(
        "openai.AsyncOpenAI",
        return_value=MockAsyncOpenAI(
            remaining_run_status=["completed"], remaining_messages=["welcome"]
        ),
    )
    async def test_begin_task_single_reply(self, mock_async_openai):
        context = self.create_mock_context()
        state = await self.create_state(context)
        state.temp.input = "hello"

        assistants_planner = AssistantsPlanner(
            OpenAIAssistantsOptions(
                api_key="test-key",
                assistant_id=ASSISTANT_ID,
            )
        )

        plan = await assistants_planner.begin_task(context, state)

        self.assertTrue(mock_async_openai.called)
        self.assertNotEqual(plan, None)
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "SAY")
        commands = cast(List[PredictedSayCommand], plan.commands)
        if commands[0].response is not None:
            self.assertEqual("welcome", commands[0].response.content)
        else:
            self.fail("commands[0].response is None")

    @mock.patch(
        "openai.AsyncOpenAI",
        return_value=MockAsyncOpenAI(
            remaining_run_status=["in_progress", "completed"], remaining_messages=["welcome"]
        ),
    )
    async def test_begin_task_waits_for_run(self, mock_async_openai):
        context = self.create_mock_context()
        state = await self.create_state(context)
        state.temp.input = "hello"

        assistants_planner = AssistantsPlanner(
            OpenAIAssistantsOptions(
                api_key="test-key",
                assistant_id=ASSISTANT_ID,
            )
        )

        plan = await assistants_planner.begin_task(context, state)

        self.assertTrue(mock_async_openai.called)
        self.assertNotEqual(plan, None)
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "SAY")
        commands = cast(List[PredictedSayCommand], plan.commands)
        assert commands[0].response is not None
        self.assertEqual("welcome", commands[0].response.content)

    @mock.patch(
        "openai.AsyncOpenAI",
        return_value=MockAsyncOpenAI(
            shared_messages={"123": []},
            shared_threads=[Thread(id="123", created_at=12323, object="thread")],
            shared_runs={
                "123": [
                    Run(
                        id="234",
                        thread_id="123",
                        assistant_id=ASSISTANT_ID,
                        status="in_progress",
                        created_at=233,
                        expires_at=23123,
                        instructions=ASSISTANT_INSTRUCTIONS,
                        model=ASSISTANT_MODEL,
                        object="thread.run",
                        tools=ASSISTANT_TOOLS,
                    )
                ]
            },
            remaining_run_status=["failed", "completed"],
            remaining_messages=["welcome"],
        ),
    )
    async def test_begin_task_waits_for_previous_run(self, mock_async_openai):
        context = self.create_mock_context()
        state = await self.create_state(context)
        state.temp.input = "hello"

        assistants_planner = AssistantsPlanner(
            OpenAIAssistantsOptions(
                api_key="test-key",
                assistant_id=ASSISTANT_ID,
            )
        )

        state.set("conversation.assistants_state", {"thread_id": "123"})

        plan = await assistants_planner.begin_task(context, state)

        self.assertTrue(mock_async_openai.called)
        self.assertNotEqual(plan, None)
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "SAY")
        commands = cast(List[PredictedSayCommand], plan.commands)
        assert commands[0].response is not None
        self.assertEqual("welcome", commands[0].response.content)

    @mock.patch(
        "openai.AsyncOpenAI",
        return_value=MockAsyncOpenAI(
            remaining_run_status=["cancelled"], remaining_messages=["welcome"]
        ),
    )
    async def test_begin_task_run_cancelled(self, mock_async_openai):
        context = self.create_mock_context()
        state = await self.create_state(context)
        state.temp.input = "hello"

        assistants_planner = AssistantsPlanner(
            OpenAIAssistantsOptions(
                api_key="test-key",
                assistant_id=ASSISTANT_ID,
            )
        )

        plan = await assistants_planner.begin_task(context, state)

        self.assertTrue(mock_async_openai.called)
        self.assertNotEqual(plan, None)
        self.assertEqual(len(plan.commands), 0)

    @mock.patch(
        "openai.AsyncOpenAI",
        return_value=MockAsyncOpenAI(
            remaining_run_status=["expired"], remaining_messages=["welcome"]
        ),
    )
    async def test_begin_task_run_expired(self, mock_async_openai):
        context = self.create_mock_context()
        state = await self.create_state(context)
        state.temp.input = "hello"

        assistants_planner = AssistantsPlanner(
            OpenAIAssistantsOptions(
                api_key="test-key",
                assistant_id=ASSISTANT_ID,
            )
        )

        plan = await assistants_planner.begin_task(context, state)

        self.assertTrue(mock_async_openai.called)
        self.assertNotEqual(plan, None)
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "DO")
        self.assertEqual(
            ActionTypes.TOO_MANY_STEPS, cast(PredictedDoCommand, plan.commands[0]).action
        )

    @mock.patch(
        "openai.AsyncOpenAI",
        return_value=MockAsyncOpenAI(
            remaining_run_status=["failed"], remaining_messages=["welcome"]
        ),
    )
    async def test_begin_task_run_failed(self, mock_async_openai):
        context = self.create_mock_context()
        state = await self.create_state(context)
        state.temp.input = "hello"

        assistants_planner = AssistantsPlanner(
            OpenAIAssistantsOptions(
                api_key="test-key",
                assistant_id=ASSISTANT_ID,
            )
        )

        self.assertTrue(mock_async_openai.called)

        with self.assertRaises(ApplicationError):
            await assistants_planner.begin_task(context, state)

    @mock.patch(
        "openai.AsyncOpenAI",
        return_value=MockAsyncOpenAI(
            remaining_actions=[
                run.RequiredAction(
                    type="submit_tool_outputs",
                    submit_tool_outputs=run.RequiredActionSubmitToolOutputs(
                        tool_calls=[
                            RequiredActionFunctionToolCall(
                                type="function",
                                id="test-tool-id",
                                function=required_action_function_tool_call.Function(
                                    arguments="{}", name="test-action"
                                ),
                            )
                        ]
                    ),
                )
            ],
            remaining_run_status=["requires_action", "in_progress", "completed"],
            remaining_messages=["welcome"],
        ),
    )
    async def test_continue_task_requires_action(self, mock_async_openai):
        context = self.create_mock_context()
        state = await self.create_state(context)
        state.temp.input = "hello"

        assistants_planner = AssistantsPlanner(
            OpenAIAssistantsOptions(
                api_key="test-key",
                assistant_id=ASSISTANT_ID,
            )
        )

        plan1 = await assistants_planner.continue_task(context, state)
        state.temp.action_outputs["test-action"] = "test-output"
        plan2 = await assistants_planner.continue_task(context, state)
        commands1 = cast(List[PredictedDoCommand], plan1.commands)
        commands2 = cast(List[PredictedSayCommand], plan2.commands)

        self.assertTrue(mock_async_openai.called)
        self.assertNotEqual(plan1, None)
        self.assertEqual(len(plan1.commands), 1)
        self.assertEqual(plan1.commands[0].type, "DO")
        if commands1[0].action is not None:
            self.assertEqual("test-action", commands1[0].action)
        else:
            self.fail("commands1[0].action is None")

        self.assertNotEqual(plan2, None)
        self.assertEqual(len(plan2.commands), 1)
        self.assertEqual(plan2.commands[0].type, "SAY")
        if commands2[0].response is not None:
            self.assertEqual("welcome", commands2[0].response.content)
        else:
            self.fail("commands2[0].response is None")

        tool_map = state.get("temp.submit_tool_map")
        if tool_map:
            self.assertTrue("test-action" in tool_map)
            self.assertEqual(tool_map["test-action"], "test-tool-id")

    @mock.patch(
        "openai.AsyncOpenAI",
        return_value=MockAsyncOpenAI(
            remaining_actions=[
                run.RequiredAction(
                    type="submit_tool_outputs",
                    submit_tool_outputs=run.RequiredActionSubmitToolOutputs(
                        tool_calls=[
                            RequiredActionFunctionToolCall(
                                type="function",
                                id="test-tool-id",
                                function=required_action_function_tool_call.Function(
                                    arguments="{}", name="test-action"
                                ),
                            )
                        ]
                    ),
                )
            ],
            remaining_run_status=["requires_action", "in_progress", "completed"],
            remaining_messages=["welcome"],
        ),
    )
    async def test_continue_task_ignores_redundant_action(self, mock_async_openai):
        context = self.create_mock_context()
        state = await self.create_state(context)
        state.temp.input = "hello"
        state.temp.action_outputs["other-action"] = "should not be used"

        assistants_planner = AssistantsPlanner(
            OpenAIAssistantsOptions(
                api_key="test-key",
                assistant_id=ASSISTANT_ID,
            )
        )

        plan1 = await assistants_planner.continue_task(context, state)
        state.temp.action_outputs["test-action"] = "test-output"
        plan2 = await assistants_planner.continue_task(context, state)

        self.assertTrue(mock_async_openai.called)
        self.assertNotEqual(plan1, None)
        self.assertEqual(len(plan1.commands), 1)
        self.assertEqual(plan1.commands[0].type, "DO")
        self.assertEqual("test-action", cast(PredictedDoCommand, plan1.commands[0]).action)

        self.assertNotEqual(plan2, None)
        self.assertEqual(len(plan2.commands), 1)
        self.assertEqual(plan2.commands[0].type, "SAY")
        commands = cast(List[PredictedSayCommand], plan2.commands)
        assert commands[0].response is not None
        self.assertEqual("welcome", commands[0].response.content)

        tool_map = state.get("temp.submit_tool_map")
        if tool_map:
            self.assertTrue("test-action" in tool_map)
            self.assertEqual(tool_map["test-action"], "test-tool-id")

    @mock.patch(
        "openai.AsyncOpenAI",
        return_value=MockAsyncOpenAI(
            remaining_run_status=["completed"],
            remaining_messages=["message 2", "message 1", "welcome"],
        ),
    )
    async def test_continue_task_multiple_messages(self, mock_async_openai):
        context = self.create_mock_context()
        state = await self.create_state(context)
        state.temp.input = "hello"
        state.temp.action_outputs["other-action"] = "should not be used"

        assistants_planner = AssistantsPlanner(
            OpenAIAssistantsOptions(
                api_key="test-key",
                assistant_id=ASSISTANT_ID,
            )
        )

        plan = await assistants_planner.continue_task(context, state)
        commands = cast(List[PredictedSayCommand], plan.commands)

        self.assertTrue(mock_async_openai.called)
        self.assertNotEqual(plan, None)
        self.assertEqual(len(commands), 3)

        self.assertEqual(commands[0].type, "SAY")
        assert commands[0].response is not None
        self.assertEqual("message 2", commands[0].response.content)

        assert commands[1].type == "SAY"
        assert commands[1].response is not None
        self.assertEqual("message 1", commands[1].response.content)

        assert commands[2].type == "SAY"
        assert commands[2].response is not None
        self.assertEqual("welcome", commands[2].response.content)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Literal

ActivityType = Literal[
    "message",
    "contactRelationUpdate",
    "conversationUpdate",
    "typing",
    "endOfConversation",
    "event",
    "invoke",
    "invokeResponse",
    "deleteUserData",
    "messageUpdate",
    "messageDelete",
    "installationUpdate",
    "messageReaction",
    "suggestion",
    "trace",
    "handoff",
    "command",
    "commandResult",
]

ConversationUpdateType = Literal[
    "channelCreated",
    "channelRenamed",
    "channelDeleted",
    "channelRestored",
    "membersAdded",
    "membersRemoved",
    "teamRenamed",
    "teamDeleted",
    "teamArchived",
    "teamUnarchived",
    "teamRestored",
]

MessageReactionType = Literal["reactionsAdded", "reactionsRemoved"]

MessageUpdateType = Literal["editMessage", "softDeleteMessage", "undeleteMessage"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from enum import Enum


class MessageReactionTypes(str, Enum):
    LIKE = "like"
    PLUS_ONE = "plusOne"



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar

from botbuilder.core import TurnContext

from ...state import TurnState
from ...user_agent import _UserAgent
from ..planners.plan import Plan

StateT = TypeVar("StateT", bound=TurnState)


class Moderator(ABC, _UserAgent, Generic[StateT]):
    """
    A moderator is responsible for reviewing and approving AI prompts and plans.
    """

    @abstractmethod
    async def review_input(self, context: TurnContext, state: StateT) -> Optional[Plan]:
        """
        Reviews an incoming utterance and generated prompt before it's sent to the planner.

        Args:
            context (TurnContext): the current turn of conversation.
            state (TurnState): the current turn state.

        Returns:
            Optional[Plan]: if approved then returns None, otherwise returns a new plan
                to redirect.
        """

    @abstractmethod
    async def review_output(self, context: TurnContext, state: StateT, plan: Plan) -> Plan:
        """
        Reviews a plan generated by the planner before its executed.

        Args:
            context (TurnContext): the current turn of conversation.
            state (TurnState): the current turn state.
            plan (Plan): plan generated by the planner.

        Returns:
            Plan: the plan to execute. Either the current plan passed in for review
                or a new plan.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List

from botbuilder.core import TurnContext

from ....state import MemoryBase
from ...tokenizers import Tokenizer
from ..message import Message
from ..prompt_functions import PromptFunctions
from ..rendered_prompt_section import RenderedPromptSection


class PromptSection(ABC):
    """
    A section that can be rendered to a prompt as either text or an array of `Message` objects.
    """

    @property
    @abstractmethod
    def required(self) -> bool:
        """
        If true the section is mandatory otherwise it can be safely dropped.
        """

    @property
    @abstractmethod
    def tokens(self) -> float:
        """
        The requested token budget for this section.
        - Values between 0.0 and 1.0 represent a percentage of the total budget and
          the section will be layed out proportionally to all other sections.
        - Values greater than 1.0 represent the max number of tokens the section
          should be allowed to consume.
        """

    @abstractmethod
    async def render_as_text(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> "RenderedPromptSection[str]":
        """
        Renders the section as a string of text.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[str]: The rendered prompt section as a string.
        """

    @abstractmethod
    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> "RenderedPromptSection[List[Message]]":
        """
        Renders the section as a list of messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered prompt section as a list of messages.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations


class ApplicationError(Exception):
    """
    Teams Application Error
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase

import pytest
from botbuilder.core import TurnContext

from teams.ai.data_sources import TextDataSource
from teams.ai.tokenizers import GPTTokenizer, Tokenizer
from teams.state import TurnState


class TestTextDataSource(IsolatedAsyncioTestCase):
    tokenizer: Tokenizer

    @pytest.fixture(autouse=True)
    def before_each(self):
        self.tokenizer = GPTTokenizer()
        yield

    def setUp(self):
        self.text_data_source = TextDataSource("testname", "Hello World!")

    async def test_render_data_returns_trimmed_text(self):
        state = TurnState()
        section = await self.text_data_source.render_data(
            turn_context=cast(TurnContext, {}), memory=state, tokenizer=self.tokenizer, max_tokens=1
        )

        self.assertEqual(section.output, "Hello")
        self.assertEqual(section.too_long, True)
        self.assertEqual(section.length, 1)

    async def test_render_data_returns_full_text(self):
        state = TurnState()
        section = await self.text_data_source.render_data(
            turn_context=cast(TurnContext, {}),
            memory=state,
            tokenizer=self.tokenizer,
            max_tokens=100,
        )

        self.assertEqual(section.output, "Hello World!")
        self.assertEqual(section.too_long, False)
        self.assertEqual(section.length, 3)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import TypeVar, Union

from botbuilder.core import TurnContext

from ...state import MemoryBase
from ..models.prompt_response import PromptResponse
from ..planners.plan import Plan, PredictedSayCommand
from ..prompts.sections.prompt_section import PromptSection
from ..tokenizers.tokenizer import Tokenizer
from ..validators.validation import Validation
from .augmentation import Augmentation

ContentT = TypeVar("ContentT")


class DefaultAugmentation(Augmentation[str]):
    """
    The default 'none' augmentation.

    This augmentation does not add any additional functionality to the prompt. It always
    returns a `Plan` with a single `SAY` command containing the models response.
    """

    def create_prompt_section(self) -> Union[PromptSection, None]:
        """
        Creates an optional prompt section for the augmentation.
        """
        return None

    async def validate_response(
        self,
        context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        response: PromptResponse[str],
        remaining_attempts: int,
    ) -> Validation:
        """
        Validates a response to a prompt.

        Args:
            context (TurnContext): Context for the current turn of conversation.
            memory (MemoryBase): Interface for accessing state variables.
            tokenizer (Tokenizer): Tokenizer to use for encoding/decoding text.
            response (PromptResponse[str]): Response to validate.
            remaining_attempts (int): Nubmer of remaining attempts to validate the response.

        Returns:
            Validation: A 'Validation' object.
        """
        return Validation(valid=True)

    async def create_plan_from_response(
        self, turn_context: TurnContext, memory: MemoryBase, response: PromptResponse[str]
    ) -> Plan:
        """
        Creates a plan given validated response value.

        Args:
            turn_context (TurnContext): Context for the current turn of conversation.
            memory (MemoryBase): Interface for accessing state variables.
            response (PromptResponse[str]): The validated and transformed response for the prompt.

        Returns:
            Plan: The created plan.
        """
        say_response = response.message
        return Plan(commands=[PredictedSayCommand(response=say_response)])



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .auth import Auth
from .auth_manager import AuthManager
from .auth_options import AuthOptions
from .oauth import OAuth, OAuthDialog, OAuthOptions
from .sign_in_response import AuthErrorReason, SignInResponse, SignInStatus

__all__ = [
    "SignInStatus",
    "SignInResponse",
    "Auth",
    "AuthManager",
    "AuthOptions",
    "AuthErrorReason",
    "OAuth",
    "OAuthOptions",
    "OAuthDialog",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, MagicMock

from botbuilder.core import MemoryStorage, TurnContext

from teams.ai.prompts import PromptFunctions, TemplateSection
from teams.ai.tokenizers import GPTTokenizer
from teams.app_error import ApplicationError
from teams.state import ConversationState, TempState, TurnState, UserState


class TestTemplateSection(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.context = MagicMock(spec=TurnContext)
        self.context.activity.channel_id = "channel_id"
        self.context.activity.recipient.id = "bot_id"
        self.context.activity.conversation.id = "conversation_id"
        self.context.activity.from_property.id = "user_id"

        memory_storage = MemoryStorage()
        self.memory = await TurnState[ConversationState, UserState, TempState].load(
            self.context, memory_storage
        )
        self.memory.conversation["test_property"] = "conversation_test_value"
        self.memory.user["test_property"] = "user_test_value"
        self.memory.temp.input = "temp_input"

        self.functions = MagicMock(spec=PromptFunctions)
        mock_invoke_function = AsyncMock()
        mock_invoke_function.return_value = "test_func"
        self.functions.invoke_function = mock_invoke_function
        self.tokenizer = GPTTokenizer()

    async def test_init(self):
        template_section = TemplateSection("template", "user")
        self.assertEqual(template_section.template, "template")
        self.assertEqual(template_section.role, "user")
        self.assertEqual(template_section.tokens, -1)
        self.assertTrue(template_section.required)
        self.assertEqual(template_section.separator, "\n")
        self.assertEqual(template_section.text_prefix, "")

    async def test_init_with_optional_params(self):
        template_section = TemplateSection("template", "user", 10, False, " ", "text_prefix")
        self.assertEqual(template_section.template, "template")
        self.assertEqual(template_section.role, "user")
        self.assertEqual(template_section.tokens, 10)
        self.assertFalse(template_section.required)
        self.assertEqual(template_section.separator, " ")
        self.assertEqual(template_section.text_prefix, "text_prefix")

    async def test_init_with_missing_brace(self):
        with self.assertRaises(ApplicationError) as context:
            TemplateSection("Hello {{test_func}", "user")  # missing closing brace
        self.assertEqual(str(context.exception), "Invalid template: Hello {{test_func}")

    async def test_init_with_with_unclosed_function_arg(self):
        with self.assertRaises(ApplicationError) as context:
            TemplateSection("Hello {{test_func 'arg1}}", "role")
        self.assertEqual(str(context.exception), "Invalid template: Hello {{test_func 'arg1}}")

    async def test_render_as_messages(self):
        template_section = TemplateSection("Hello World", "role")
        result = await template_section.render_as_messages(
            self.context, self.memory, self.functions, self.tokenizer, 10
        )
        self.assertEqual(result.output[0].role, "role")
        self.assertEqual(result.output[0].content, "Hello World")
        self.assertFalse(result.too_long)
        self.assertEqual(result.length, 2)

    async def test_render_as_messages_too_long(self):
        template_section = TemplateSection("Hello World", "role")
        result = await template_section.render_as_messages(
            self.context, self.memory, self.functions, self.tokenizer, 1
        )
        self.assertEqual(result.output[0].role, "role")
        self.assertEqual(result.output[0].content, "Hello World")
        self.assertTrue(result.too_long)
        self.assertEqual(result.length, 2)

    async def test_render_as_text(self):
        template_section = TemplateSection("Hello World", "role")
        result = await template_section.render_as_text(
            self.context, self.memory, self.functions, self.tokenizer, 10
        )
        self.assertEqual(result.output, "Hello World")
        self.assertFalse(result.too_long)
        self.assertEqual(result.length, 2)

    async def test_render_as_text_too_long(self):
        template_section = TemplateSection("Hello World", "role")
        result = await template_section.render_as_text(
            self.context, self.memory, self.functions, self.tokenizer, 1
        )
        self.assertEqual(result.output, "Hello World")
        self.assertTrue(result.too_long)
        self.assertEqual(result.length, 2)

    async def test_render_variable(self):
        template_section = TemplateSection(
            "Hello {{$user.test_property}} {{$conversation.test_property}} {{$temp.input}}", "role"
        )
        result = await template_section.render_as_text(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        self.assertEqual(
            result.output, 'Hello "user_test_value" "conversation_test_value" "temp_input"'
        )
        self.assertEqual(result.length, 15)
        self.assertFalse(result.too_long)

    async def test_render_variable_with_whitespace(self):
        template_section = TemplateSection(
            "Hello {{ $user.test_property }} {{ $conversation.test_property }} {{ $temp.input }}",
            "role",
        )
        result = await template_section.render_as_text(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        self.assertEqual(
            result.output, 'Hello "user_test_value" "conversation_test_value" "temp_input"'
        )
        self.assertEqual(result.length, 15)
        self.assertFalse(result.too_long)

    async def test_render_non_exist_variable(self):
        template_section = TemplateSection("Hello {{$user.non_exist}}", "role")
        result = await template_section.render_as_text(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        self.assertEqual(result.output, "Hello ")
        self.assertEqual(result.length, 2)
        self.assertFalse(result.too_long)

    async def test_render_function(self):
        template_section = TemplateSection("Hello {{test_func}}", "role")
        result = await template_section.render_as_text(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        args, _kwargs = self.functions.invoke_function.call_args
        name_param_value = args[0]
        args_parama_value = args[4]
        self.assertEqual(result.output, 'Hello "test_func"')
        self.assertEqual(name_param_value, "test_func")
        self.assertEqual(args_parama_value, [])
        self.assertEqual(result.length, 5)
        self.assertFalse(result.too_long)

    async def test_render_function_with_args(self):
        template_section = TemplateSection("Hello {{test_func arg1 arg2}}", "role")
        result = await template_section.render_as_text(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        args, _kwargs = self.functions.invoke_function.call_args
        name_param_value = args[0]
        args_parama_value = args[4]
        self.assertEqual(result.output, 'Hello "test_func"')
        self.assertEqual(name_param_value, "test_func")
        self.assertEqual(args_parama_value, ["arg1", "arg2"])
        self.assertEqual(result.length, 5)
        self.assertFalse(result.too_long)

    async def test_render_function_with_args_and_whitespaces(self):
        template_section = TemplateSection("Hello {{ test_func arg1 arg2 }}", "role")
        result = await template_section.render_as_text(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        args, _kwargs = self.functions.invoke_function.call_args
        name_param_value = args[0]
        args_parama_value = args[4]
        self.assertEqual(result.output, 'Hello "test_func"')
        self.assertEqual(name_param_value, "test_func")
        self.assertEqual(args_parama_value, ["arg1", "arg2"])
        self.assertEqual(result.length, 5)
        self.assertFalse(result.too_long)

    async def test_render_function_with_quoted_args(self):
        template_section = TemplateSection("Hello {{ test_func 'arg1' \"arg2\" `arg3` }}", "role")
        result = await template_section.render_as_text(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        args, _kwargs = self.functions.invoke_function.call_args
        name_param_value = args[0]
        args_parama_value = args[4]
        self.assertEqual(result.output, 'Hello "test_func"')
        self.assertEqual(name_param_value, "test_func")
        self.assertEqual(args_parama_value, ["arg1", "arg2", "arg3"])
        self.assertEqual(result.length, 5)
        self.assertFalse(result.too_long)

    async def test_render_empty_block(self):
        template_section = TemplateSection("Hello {{}}", "role")
        result = await template_section.render_as_text(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        self.assertEqual(result.output, "Hello ")
        self.assertEqual(result.length, 2)
        self.assertFalse(result.too_long)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .assistant_message import AssistantMessage
from .augmentation_config import AugmentationConfig
from .completion_config import CompletionConfig
from .function_call import FunctionCall
from .function_call_message import FunctionCallMessage
from .function_response_message import FunctionResponseMessage
from .message import (
    Citation,
    ImageContentPart,
    ImageUrl,
    Message,
    MessageContext,
    TextContentPart,
)
from .prompt import Prompt
from .prompt_functions import PromptFunction, PromptFunctions
from .prompt_manager import PromptManager
from .prompt_manager_options import PromptManagerOptions
from .prompt_template import PromptTemplate
from .prompt_template_config import PromptTemplateConfig
from .rendered_prompt_section import RenderedPromptSection
from .sections import *
from .system_message import SystemMessage
from .user_input_message import UserInputMessage
from .user_message import UserMessage

__all__ = [
    "AssistantMessage",
    "AugmentationConfig",
    "CompletionConfig",
    "Citation",
    "FunctionCall",
    "FunctionCallMessage",
    "FunctionResponseMessage",
    "ImageContentPart",
    "ImageUrl",
    "Message",
    "MessageContext",
    "TextContentPart",
    "Prompt",
    "PromptFunction",
    "PromptFunctions",
    "PromptManager",
    "PromptManagerOptions",
    "PromptTemplate",
    "PromptTemplateConfig",
    "RenderedPromptSection",
    "SystemMessage",
    "UserInputMessage",
    "UserMessage",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .action_planner import (
    ActionPlanner,
    ActionPlannerOptions,
    ActionPlannerPromptFactory,
)
from .assistants_planner import (
    AssistantsPlanner,
    AzureOpenAIAssistantsOptions,
    OpenAIAssistantsOptions,
)
from .plan import Plan, PredictedCommand, PredictedDoCommand, PredictedSayCommand
from .planner import Planner

__all__ = [
    "ActionPlanner",
    "ActionPlannerOptions",
    "AssistantsPlanner",
    "OpenAIAssistantsOptions",
    "AzureOpenAIAssistantsOptions",
    "ActionPlannerPromptFactory",
    "Plan",
    "PredictedCommand",
    "PredictedDoCommand",
    "PredictedSayCommand",
    "Planner",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase

from botbuilder.core import TurnContext

from teams.ai.moderators import DefaultModerator
from teams.ai.planners import Plan
from teams.state import TurnState


class TestDefaultModerator(IsolatedAsyncioTestCase):
    moderator = DefaultModerator()

    async def test_review_input(self):
        plan = await self.moderator.review_input(context=cast(TurnContext, {}), state=TurnState())

        self.assertIsNone(plan)

    async def test_review_output(self):
        plan = Plan()
        output = await self.moderator.review_output(
            context=cast(TurnContext, {}), state=TurnState(), plan=plan
        )

        self.assertEqual(output, plan)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations


def snippet(text: str, max_length: int) -> str:
    """
    Clips the text to a maximum length in case it exceeds the limit.

    Args:
        text: str The text to clip.
        maxLength The maximum length of the text to return, cutting off the last whole word.

    Returns:
        str: The modified text
    """
    if len(text) <= max_length:
        return text
    snippet = text[:max_length]
    snippet = snippet[: max(snippet.rfind(" "), -1)]
    snippet += "..."
    return snippet



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from botbuilder.core import TurnContext
from dataclasses_json import DataClassJsonMixin, dataclass_json

from ....state import MemoryBase
from ...models.chat_completion_action import ChatCompletionAction
from ...tokenizers.tokenizer import Tokenizer
from ..message import Message
from ..prompt_functions import PromptFunctions
from ..rendered_prompt_section import RenderedPromptSection
from .prompt_section_base import PromptSectionBase


@dataclass_json
@dataclass
class ActionValue(DataClassJsonMixin):
    description: Optional[str] = None
    parameters: Optional[Union[Dict[str, Any], Dict[str, Dict[str, Any]]]] = None


@dataclass_json
@dataclass
class ActionList(DataClassJsonMixin):
    actions: Dict[str, ActionValue]


class ActionAugmentationSection(PromptSectionBase):
    """
    A prompt section that renders a list of actions to the prompt.
    """

    _text: str
    _token_list: Optional[List[int]] = None
    _actions: Dict[str, ChatCompletionAction] = {}

    @property
    def actions(self) -> Dict[str, ChatCompletionAction]:
        """
        Map of action names to actions.
        """
        return self._actions

    def __init__(self, actions: List[ChatCompletionAction], call_to_action: str) -> None:
        """
        Creates a new `ActionAugmentationSection` instance.

        Args:
            actions (List[ChatCompletionAction]): List of actions to render.
            call_to_action (str): Text to display after the list of actions.

        """
        super().__init__(-1, True, "\n\n")

        # Convert actions to an ActionList
        action_list = ActionList(actions={})

        for action in actions:
            self._actions[action.name] = action
            action_list.actions[action.name] = ActionValue(description=action.description)
            if action.parameters:
                params = action.parameters
                updated_params = (
                    params.get("properties")
                    if params.get("additional_properties") is None
                    else params
                )
                action_list.actions[action.name] = ActionValue(
                    description=action.description, parameters=updated_params
                )

        # Build augmentation text
        self._text = f"{action_list.to_json()}\n\n{call_to_action}"

    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message[str]]]:
        """
        Renders the prompt section as a list of `Message` objects.

        Args:
            context (TurnContext): Context for the current turn of conversation.
            memory (MemoryBase): Interface for accessing state variables.
            functions (PromptFunctions): Functions for rendering prompts.
            tokenizer (Tokenizer): Tokenizer to use for encoding/decoding text.
            max_tokens (int): Maximum number of tokens allowed for the rendered prompt.

        Returns:
            RenderedPromptSection[List[Message[str]]]: The rendered prompt section.

        """
        # Tokenize on first use
        if not self._token_list:
            self._token_list = tokenizer.encode(self._text)

        # Check for max tokens
        if len(self._token_list) > max_tokens:
            trimmed = self._token_list[0:max_tokens]
            return RenderedPromptSection[List[Message[str]]](
                output=[Message[str](role="system", content=tokenizer.decode(trimmed))],
                length=len(trimmed),
                too_long=True,
            )
        return RenderedPromptSection[List[Message[str]]](
            output=[Message[str](role="system", content=self._text)],
            length=len(self._token_list),
            too_long=False,
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Optional, cast

from botbuilder.dialogs import OAuthPromptSettings
from botframework.connector.auth import AppCredentials


class OAuthOptions(OAuthPromptSettings):
    "The settings for OAuthAuthentication."

    enable_sso: bool = False
    "Set to `true` to enable SSO when authenticating using Azure Active Directory (AAD)."

    token_exchange_url: Optional[str] = None
    "The token exchange uri for SSO in adaptive card auth scenario."

    def __init__(
        self,
        *,
        connection_name: str,
        title: str,
        text: Optional[str] = None,
        timeout: Optional[int] = None,
        oauth_app_credentials: Optional[AppCredentials] = None,
        end_on_invalid_message: bool = False,
        enable_sso: bool = False,
        token_exchange_url: Optional[str] = None,
    ) -> None:
        super().__init__(
            connection_name,
            title,
            text or title,
            cast(int, timeout),
            cast(AppCredentials, oauth_app_credentials),
            end_on_invalid_message,
        )

        self.enable_sso = enable_sso
        self.token_exchange_url = token_exchange_url



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from logging import Logger
from typing import Any, Awaitable, Callable, Optional, cast

from aiohttp import ClientResponse, ClientResponseError, ClientSession
from aiohttp.web import Request, Response, WebSocketResponse
from botbuilder.core import Bot
from botbuilder.core.turn_context import TurnContext
from botbuilder.integration.aiohttp import (
    CloudAdapter,
    ConfigurationBotFrameworkAuthentication,
    ConfigurationServiceClientCredentialFactory,
)
from botbuilder.schema import Activity
from botframework.connector import (
    ConnectorClient,
    HttpClientBase,
    HttpRequest,
    HttpResponseBase,
)
from botframework.connector.auth import (
    AuthenticationConfiguration,
    ClaimsIdentity,
    HttpClientFactory,
    ServiceClientCredentialsFactory,
)
from botframework.connector.auth.connector_factory import ConnectorFactory
from botframework.connector.auth.user_token_client import UserTokenClient

from .user_agent import _UserAgent


class TeamsAdapter(CloudAdapter, _UserAgent):
    """
    An adapter that implements the Bot Framework Protocol
    and can be hosted in different cloud environments both public and private.
    """

    _credentials_factory: ServiceClientCredentialsFactory
    "The credentials factory used by the bot adapter to create a [ServiceClientCredentials] object."
    _configuration: Any

    @property
    def credentials_factory(self) -> ServiceClientCredentialsFactory:
        """
        The bot's credentials factory.
        """

        return self._credentials_factory

    @property
    def configuration(self) -> Any:
        """
        The bot's configuration.
        """

        return self._configuration

    def __init__(
        self,
        configuration: Any,
        *,
        credentials_factory: Optional[ServiceClientCredentialsFactory] = None,
        auth_configuration: Optional[AuthenticationConfiguration] = None,
        http_client_factory: Optional[HttpClientFactory] = None,
        logger: Optional[Logger] = None,
    ) -> None:
        """
        Initializes a new instance of the TeamsAdapter class.
        """

        super().__init__(
            ConfigurationBotFrameworkAuthentication(
                configuration,
                credentials_factory=cast(ServiceClientCredentialsFactory, credentials_factory),
                auth_configuration=cast(AuthenticationConfiguration, auth_configuration),
                http_client_factory=_TeamsHttpClientFactory(parent=http_client_factory),
                logger=cast(Logger, logger),
            )
        )

        self._configuration = configuration
        self._credentials_factory = (
            cast(ServiceClientCredentialsFactory, credentials_factory)
            if credentials_factory
            else ConfigurationServiceClientCredentialFactory(configuration)
        )

    async def process(
        self,
        request: Request,
        bot: Bot,
        ws_response: Optional[WebSocketResponse] = None,
    ) -> Optional[Response]:
        res = await super().process(
            request,
            bot,
            cast(WebSocketResponse, ws_response),
        )

        if res is not None:
            res.headers.add("User-Agent", self.user_agent)

        return res

    def _create_turn_context(
        self,
        activity: Activity,
        claims_identity: ClaimsIdentity,
        oauth_scope: str,
        connector_client: ConnectorClient,
        user_token_client: UserTokenClient,
        logic: Callable[[TurnContext], Awaitable],
        connector_factory: ConnectorFactory,
    ) -> TurnContext:
        connector_client.config.add_user_agent(self.user_agent)
        connector_client.conversations.config.add_user_agent(self.user_agent)
        connector_client.attachments.config.add_user_agent(self.user_agent)

        return super()._create_turn_context(
            activity,
            claims_identity,
            oauth_scope,
            connector_client,
            user_token_client,
            logic,
            connector_factory,
        )


class _TeamsHttpClientFactory(HttpClientFactory):
    _parent: Optional[HttpClientFactory]

    def __init__(self, *, parent: Optional[HttpClientFactory] = None) -> None:
        self._parent = parent

    def create_client(self) -> HttpClientBase:
        if self._parent:
            return _TeamsHttpClient(parent=self._parent.create_client())

        return _TeamsHttpClient()


class _TeamsHttpClient(HttpClientBase, _UserAgent):
    _parent: Optional[HttpClientBase]

    def __init__(self, *, parent: Optional[HttpClientBase] = None) -> None:
        self._session = ClientSession()
        self._parent = parent

    async def post(self, *, request: HttpRequest) -> HttpResponseBase:
        request.headers["User-Agent"] = self.user_agent

        if self._parent:
            return await self._parent.post(request=request)

        aio_response = await self._session.post(
            request.request_uri,
            data=request.content,
            headers=request.headers,
        )

        return _TeamsHttpResponse(aio_response)


class _TeamsHttpResponse(HttpResponseBase):
    def __init__(self, client_response: ClientResponse) -> None:
        self._client_response = client_response

    @property
    def status_code(self):
        return self._client_response.status

    async def is_succesful(self) -> bool:
        try:
            self._client_response.raise_for_status()
            return True
        except ClientResponseError:
            return False

    async def read_content_str(self) -> str:
        return (await self._client_response.read()).decode()



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase, mock

import pytest
from botbuilder.core import TurnContext
from botbuilder.schema import Activity, ChannelAccount, ConversationAccount

from teams import Application
from tests.utils import SimpleAdapter


class TestAdaptiveCardss(IsolatedAsyncioTestCase):
    app: Application

    @pytest.fixture(autouse=True)
    def before_each(self):
        self.app = Application()
        yield

    @pytest.mark.asyncio
    async def test_action_execute(self):
        handler = mock.AsyncMock()
        self.app.adaptive_cards.action_execute("doStuff")(handler)
        self.assertEqual(len(self.app._routes), 1)

        # This should trigger the handler
        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                self.mock_activity(
                    type="invoke",
                    name="adaptiveCard/action",
                    value={
                        "action": {
                            "type": "Action.Execute",
                            "title": "Action.Execute",
                            "isEnabled": False,
                            "data": {"x": 123},
                            "verb": "doStuff",
                        },
                        "trigger": "manual",
                    },
                ),
            )
        )

        # This should not trigger the handler
        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                self.mock_activity(
                    type="invoke",
                    name="adaptiveCard/action",
                    value={
                        "action": {
                            "type": "Action.Execute",
                            "title": "Action.Execute",
                            "isEnabled": False,
                            "data": {"x": 123},
                            "verb": "anotherVerb",
                        },
                        "trigger": "manual",
                    },
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_action_submit(self):
        handler = mock.AsyncMock()
        self.app.adaptive_cards.action_submit("StaticSubmit")(handler)
        self.assertEqual(len(self.app._routes), 1)

        # This should trigger the handler
        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                self.mock_activity(
                    type="message",
                    value={"verb": "StaticSubmit", "choiceSelect": "visual_studio"},
                    name=None,
                ),
            )
        )

        # This should not trigger the handler
        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                self.mock_activity(
                    type="message",
                    value={"verb": "DynamicSubmit", "choiceSelect": "visual_studio"},
                    name=None,
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_search(self):
        handler = mock.AsyncMock()
        self.app.adaptive_cards.search("npmpackages")(handler)
        self.assertEqual(len(self.app._routes), 1)

        # This should trigger the handler
        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                self.mock_activity(
                    type="invoke",
                    name="application/search",
                    value={
                        "queryText": "test",
                        "queryOptions": {"skip": 0, "top": 15},
                        "dataset": "npmpackages",
                    },
                ),
            )
        )

        # This should not trigger the handler
        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                self.mock_activity(
                    type="invoke",
                    name="application/search",
                    value={
                        "queryText": "test",
                        "queryOptions": {"skip": 0, "top": 15},
                        "dataset": "nugetpackages",
                    },
                ),
            )
        )

        handler.assert_called_once()

    def mock_activity(self, type, value, name):
        return Activity(
            id="1234",
            type=type,
            name=name,
            value=value,
            from_property=ChannelAccount(id="user", name="Test User"),
            recipient=ChannelAccount(id="bot", name="Test Bot"),
            conversation=ConversationAccount(id="convo", name="Test Convo"),
            channel_id="UnitTest",
            locale="en-uS",
            service_url="https://example.org",
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .activity import *
from .adapter import *
from .test_snippet import *
from .test_to_string import *



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Any


def todict(value: Any) -> Any:
    """
    Converts a value to a
    dictionary recursively
    """

    if isinstance(value, dict):
        data = {}
        for key, val in value.items():
            if isinstance(key, str) and key.startswith("__"):
                continue

            data[key] = todict(val)
        return data

    if isinstance(value, list):
        return [todict(v) for v in value]

    if isinstance(value, object) and hasattr(value, "__dict__"):
        return todict(value.__dict__)

    return value



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .oauth import OAuth
from .oauth_adaptive_card import OAuthAdaptiveCard
from .oauth_dialog import OAuthDialog
from .oauth_message_extension import OAuthMessageExtension
from .oauth_options import OAuthOptions

__all__ = [
    "OAuth",
    "OAuthAdaptiveCard",
    "OAuthDialog",
    "OAuthMessageExtension",
    "OAuthOptions",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import re
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    List,
    Literal,
    Pattern,
    TypeVar,
    Union,
    cast,
)

from botbuilder.core import InvokeResponse, TurnContext
from botbuilder.core.serializer_helper import deserializer_helper, serializer_helper
from botbuilder.schema import Activity, ActivityTypes
from botbuilder.schema.teams import (
    AppBasedLinkQuery,
    MessagingExtensionAction,
    MessagingExtensionActionResponse,
    MessagingExtensionQuery,
    MessagingExtensionResponse,
    MessagingExtensionResult,
    TaskModuleContinueResponse,
    TaskModuleMessageResponse,
    TaskModuleResponse,
    TaskModuleTaskInfo,
)

from teams.state import TurnState

from ..route import Route

StateT = TypeVar("StateT", bound=TurnState)
MessagePreviewAction = Literal["edit", "send"]


class MessageExtensions(Generic[StateT]):
    _routes: List[Route[StateT]] = []

    def __init__(self, routes: List[Route[StateT]]) -> None:
        self._routes = routes

    def query(self, command_id: Union[str, Pattern[str]]) -> Callable[
        [
            Callable[
                [TurnContext, StateT, MessagingExtensionQuery],
                Awaitable[MessagingExtensionResult],
            ]
        ],
        Callable[
            [TurnContext, StateT, MessagingExtensionQuery], Awaitable[MessagingExtensionResult]
        ],
    ]:
        """
        Registers a handler that implements a Search based Message Extension.

        ```python
        # Use this method as a decorator
        @app.message_extensions.query("test")
        async def on_query(context: TurnContext, state: TurnState, url: str):
            return MessagingExtensionResult()

        # Pass a function to this method
        app.message_extensions.query("test")(on_query)
        ```

        #### Args:
        - `command_id`: a string or regex pattern that matches against the activities
        `command_id`
        """

        def __selector__(context: TurnContext):
            if (
                context.activity.type != ActivityTypes.invoke
                or context.activity.name != "composeExtension/query"
            ):
                return False

            return self._activity_with_command_id(context.activity, command_id)

        def __call__(
            func: Callable[
                [TurnContext, StateT, MessagingExtensionQuery],
                Awaitable[MessagingExtensionResult],
            ],
        ) -> Callable[
            [TurnContext, StateT, MessagingExtensionQuery],
            Awaitable[MessagingExtensionResult],
        ]:
            async def __invoke__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False

                value = cast(
                    MessagingExtensionQuery,
                    deserializer_helper(MessagingExtensionQuery, context.activity.value),
                )
                res = await func(context, state, value)
                await self._invoke_response(context, res)
                return True

            self._routes.append(Route[StateT](__selector__, __invoke__, True))
            return func

        return __call__

    def query_link(self, command_id: Union[str, Pattern[str]]) -> Callable[
        [Callable[[TurnContext, StateT, str], Awaitable[MessagingExtensionResult]]],
        Callable[[TurnContext, StateT, str], Awaitable[MessagingExtensionResult]],
    ]:
        """
        Registers a handler that implements a Link Unfurling based Message Extension.

        ```python
        # Use this method as a decorator
        @app.message_extensions.query_link("test")
        async def on_query_link(context: TurnContext, state: TurnState, url: str):
            return MessagingExtensionResult()

        # Pass a function to this method
        app.message_extensions.query_link("test")(on_query_link)
        ```

        #### Args:
        - `command_id`: a string or regex pattern that matches against the activities
        `command_id`
        """

        def __selector__(context: TurnContext):
            if (
                context.activity.type != ActivityTypes.invoke
                or context.activity.name != "composeExtension/queryLink"
            ):
                return False

            return self._activity_with_command_id(context.activity, command_id)

        def __call__(
            func: Callable[
                [TurnContext, StateT, str],
                Awaitable[MessagingExtensionResult],
            ],
        ) -> Callable[
            [TurnContext, StateT, str],
            Awaitable[MessagingExtensionResult],
        ]:
            async def __invoke__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False

                value = cast(
                    AppBasedLinkQuery,
                    deserializer_helper(AppBasedLinkQuery, context.activity.value),
                )

                if not isinstance(value.url, str):
                    return False

                res = await func(context, state, value.url)
                await self._invoke_response(context, res)
                return True

            self._routes.append(Route[StateT](__selector__, __invoke__, True))
            return func

        return __call__

    def anonymous_query_link(self, command_id: Union[str, Pattern[str]]) -> Callable[
        [Callable[[TurnContext, StateT, str], Awaitable[MessagingExtensionResult]]],
        Callable[[TurnContext, StateT, str], Awaitable[MessagingExtensionResult]],
    ]:
        """
        Registers a handler for a command that performs anonymous link unfurling.

        ```python
        # Use this method as a decorator
        @app.message_extensions.anonymous_query_link("test")
        async def on_anonymous_query_link(context: TurnContext, state: TurnState, url: str):
            print(url)
            return MessageExtensionResult()

        # Pass a function to this method
        app.message_extensions.anonymous_query_link("test")(on_anonymous_query_link)
        ```

        #### Args:
        - `command_id`: a string or regex pattern that matches against the activities
        `command_id`
        """

        def __selector__(context: TurnContext):
            if (
                context.activity.type != ActivityTypes.invoke
                or context.activity.name != "composeExtension/anonymousQueryLink"
            ):
                return False

            return self._activity_with_command_id(context.activity, command_id)

        def __call__(
            func: Callable[[TurnContext, StateT, str], Awaitable[MessagingExtensionResult]],
        ) -> Callable[[TurnContext, StateT, str], Awaitable[MessagingExtensionResult]]:
            async def __invoke__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False

                value = cast(
                    AppBasedLinkQuery,
                    deserializer_helper(AppBasedLinkQuery, context.activity.value),
                )

                if not isinstance(value.url, str):
                    return False

                res = await func(context, state, value.url)
                await self._invoke_response(context, res)
                return True

            self._routes.append(Route[StateT](__selector__, __invoke__, True))
            return func

        return __call__

    def query_setting_url(
        self,
    ) -> Callable[
        [Callable[[TurnContext, StateT], Awaitable[MessagingExtensionResult]]],
        Callable[[TurnContext, StateT], Awaitable[MessagingExtensionResult]],
    ]:
        """
        Registers a handler that invokes the fetch of the config settings for a Message Extension.

        ```python
        # Use this method as a decorator
        @app.message_extensions.query_setting_url()
        async def on_query_setting_url(context: TurnContext, state: TurnState):
            return MessagingExtensionResult()

        # Pass a function to this method
        app.message_extensions.query_setting_url()(on_query_setting_url)
        ```
        """

        def __selector__(context: TurnContext):
            if (
                context.activity.type != ActivityTypes.invoke
                or context.activity.name != "composeExtension/querySettingUrl"
            ):
                return False
            return True

        def __call__(
            func: Callable[
                [TurnContext, StateT],
                Awaitable[MessagingExtensionResult],
            ],
        ) -> Callable[[TurnContext, StateT], Awaitable[MessagingExtensionResult]]:
            async def __invoke__(context: TurnContext, state: StateT):
                res = await func(context, state)
                await self._invoke_action_response(context, res)
                return True

            self._routes.append(Route[StateT](__selector__, __invoke__, True))
            return func

        return __call__

    def configure_settings(
        self,
    ) -> Callable[
        [Callable[[TurnContext, StateT, Any], Awaitable[None]]],
        Callable[[TurnContext, StateT, Any], Awaitable[None]],
    ]:
        """
        Registers a handler that implements the logic to invoke
            configuring Message Extension settings

        ```python
        # Use this method as a decorator
        @app.message_extensions.configure_settings({"foo": "bar"})
        async def on_configure_settings(context: TurnContext, state: TurnState, data: Any):
            print(data)

        # Pass a function to this method
        app.message_extensions.configure_settings({"foo": "bar"})(on_configure_settings)
        ```
        """

        def __selector__(context: TurnContext):
            if (
                context.activity.type != ActivityTypes.invoke
                or context.activity.name != "composeExtension/setting"
            ):
                return False
            return True

        def __call__(
            func: Callable[[TurnContext, StateT, Any], Awaitable[None]],
        ) -> Callable[[TurnContext, StateT, Any], Awaitable[None]]:
            async def __invoke__(context: TurnContext, state: StateT):
                value = {}
                if context.activity.value:
                    value = context.activity.value

                await func(context, state, value)
                await self._invoke_response(context, MessagingExtensionResult())
                return True

            self._routes.append(Route[StateT](__selector__, __invoke__, True))
            return func

        return __call__

    def card_button_clicked(
        self,
    ) -> Callable[
        [Callable[[TurnContext, StateT, Any], Awaitable[None]]],
        Callable[[TurnContext, StateT, Any], Awaitable[None]],
    ]:
        """
        Registers a handler that implements the logic when a
            user has clicked on a button in a Message Extension card.

        ```python
        # Use this method as a decorator
        @app.message_extensions.card_button_clicked({"foo": "bar"})
        async def on_card_button_clicked(context: TurnContext, state: TurnState, data: Any):
            print(data)

        # Pass a function to this method
        app.message_extensions.card_button_clicked({"foo": "bar"})(on_card_button_clicked)
        ```
        """

        def __selector__(context: TurnContext):
            if (
                context.activity.type != ActivityTypes.invoke
                or context.activity.name != "composeExtension/onCardButtonClicked"
            ):
                return False
            return True

        def __call__(
            func: Callable[[TurnContext, StateT, Any], Awaitable[None]],
        ) -> Callable[[TurnContext, StateT, Any], Awaitable[None]]:
            async def __invoke__(context: TurnContext, state: StateT):
                value = {}
                if context.activity.value:
                    value = context.activity.value

                await func(context, state, value)
                await self._invoke_response(context, MessagingExtensionResult())
                return True

            self._routes.append(Route[StateT](__selector__, __invoke__, True))
            return func

        return __call__

    def message_preview(
        self, command_id: Union[str, Pattern[str]], action: MessagePreviewAction
    ) -> Callable[
        [
            Callable[
                [TurnContext, StateT, Activity],
                Awaitable[Union[MessagingExtensionResult, TaskModuleTaskInfo, str, None]],
            ]
        ],
        Callable[
            [TurnContext, StateT, Activity],
            Awaitable[Union[MessagingExtensionResult, TaskModuleTaskInfo, str, None]],
        ],
    ]:
        """
        Registers a handler to process an action of a message that's being
        previewed by the user prior to sending.

        ```python
        # Use this method as a decorator
        @app.message_extensions.message_preview("test", "edit")
        async def on_message_preview_edit(
            context: TurnContext,
            state: TurnState,
            activity: Activity,
        ):
            return

        # Pass a function to this method
        app.message_extensions.message_preview("test", "edit")(on_message_preview_edit)
        ```

        #### Args:
        - `command_id`: a string or regex pattern that matches against the activities
        `command_id`
        - `action`: the action to subscribe to
        """

        def __selector__(context: TurnContext):
            if (
                context.activity.type != ActivityTypes.invoke
                or context.activity.name != "composeExtension/submitAction"
                or not context.activity.value
                or not self._activity_with_command_id(context.activity, command_id)
            ):
                return False

            value = cast(
                MessagingExtensionAction,
                deserializer_helper(MessagingExtensionAction, context.activity.value),
            )
            message_preview_action = value.bot_message_preview_action

            if not isinstance(message_preview_action, str) or message_preview_action != action:
                return False

            return True

        def __call__(
            func: Callable[
                [TurnContext, StateT, Activity],
                Awaitable[Union[MessagingExtensionResult, TaskModuleTaskInfo, str, None]],
            ],
        ) -> Callable[
            [TurnContext, StateT, Activity],
            Awaitable[Union[MessagingExtensionResult, TaskModuleTaskInfo, str, None]],
        ]:
            async def __invoke__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False

                value = cast(
                    MessagingExtensionAction,
                    deserializer_helper(MessagingExtensionAction, context.activity.value),
                )

                if (
                    not isinstance(value.bot_activity_preview, list)
                    or len(value.bot_activity_preview) == 0
                ):
                    return False

                res = await func(context, state, value.bot_activity_preview[0])
                await self._invoke_action_response(context, res)
                return True

            self._routes.append(Route[StateT](__selector__, __invoke__, True))
            return func

        return __call__

    def fetch_task(self, command_id: Union[str, Pattern[str]]) -> Callable[
        [Callable[[TurnContext, StateT], Awaitable[Union[TaskModuleTaskInfo, str]]]],
        Callable[[TurnContext, StateT], Awaitable[Union[TaskModuleTaskInfo, str]]],
    ]:
        """
        Registers a handler to process the initial fetch task for an
        Action based message extension.

        ```python
        # Use this method as a decorator
        @app.message_extensions.fetch_task("test")
        async def on_fetch_task(context: TurnContext, state: TurnState):
            return TaskModuleTaskInfo()

        # Pass a function to this method
        app.message_extensions.fetch_task("test")(on_fetch_task)
        ```

        #### Args:
        - `command_id`: a string or regex pattern that matches against the activities
        `command_id`
        """

        def __selector__(context: TurnContext):
            if (
                context.activity.type != ActivityTypes.invoke
                or context.activity.name != "composeExtension/fetchTask"
            ):
                return False

            return self._activity_with_command_id(context.activity, command_id)

        def __call__(
            func: Callable[
                [TurnContext, StateT],
                Awaitable[Union[TaskModuleTaskInfo, str]],
            ],
        ) -> Callable[
            [TurnContext, StateT],
            Awaitable[Union[TaskModuleTaskInfo, str]],
        ]:
            async def __invoke__(context: TurnContext, state: StateT):
                res = await func(context, state)
                await self._invoke_task_response(context, res)
                return True

            self._routes.append(Route[StateT](__selector__, __invoke__, True))
            return func

        return __call__

    def select_item(
        self,
    ) -> Callable[
        [Callable[[TurnContext, StateT, Any], Awaitable[MessagingExtensionResult]]],
        Callable[[TurnContext, StateT, Any], Awaitable[MessagingExtensionResult]],
    ]:
        """
        Registers a handler that implements the logic to handle the
        tap actions for items returned by a Search based message extension.

        ```python
        # Use this method as a decorator
        @app.message_extensions.select_item
        async def on_select_item(context: TurnContext, state: TurnState, item: Any):
            return MessagingExtensionResult()

        # Pass a function to this method
        app.message_extensions.select_item()(on_select_item)
        ```
        """

        def __selector__(context: TurnContext):
            if (
                context.activity.type != ActivityTypes.invoke
                or context.activity.name != "composeExtension/selectItem"
            ):
                return False

            return True

        def __call__(
            func: Callable[
                [TurnContext, StateT, Any],
                Awaitable[MessagingExtensionResult],
            ],
        ) -> Callable[
            [TurnContext, StateT, Any],
            Awaitable[MessagingExtensionResult],
        ]:
            async def __invoke__(context: TurnContext, state: StateT):
                res = await func(context, state, context.activity.value)
                await self._invoke_action_response(context, res)
                return True

            self._routes.append(Route[StateT](__selector__, __invoke__, True))
            return func

        return __call__

    def submit_action(self, command_id: Union[str, Pattern[str]]) -> Callable[
        [
            Callable[
                [TurnContext, StateT, Any],
                Awaitable[Union[MessagingExtensionResult, TaskModuleTaskInfo, str, None]],
            ]
        ],
        Callable[
            [TurnContext, StateT, Any],
            Awaitable[Union[MessagingExtensionResult, TaskModuleTaskInfo, str, None]],
        ],
    ]:
        """
        Registers a handler that implements the submit action for an
        Action based Message Extension.

        ```python
        # Use this method as a decorator
        @app.message_extensions.submit_action("test")
        async def on_submit_action(context: TurnContext, state: TurnState, data: Any):
            return TaskModuleTaskInfo()

        # Pass a function to this method
        app.message_extensions.submit_action("test")(on_submit_action)
        ```

        #### Args:
        - `command_id`: a string or regex pattern that matches against the activities
        `command_id`
        """

        def __selector__(context: TurnContext):
            if (
                context.activity.type != ActivityTypes.invoke
                or context.activity.name != "composeExtension/submitAction"
            ):
                return False

            return self._activity_with_command_id(context.activity, command_id)

        def __call__(
            func: Callable[
                [TurnContext, StateT, Any],
                Awaitable[Union[MessagingExtensionResult, TaskModuleTaskInfo, str, None]],
            ],
        ) -> Callable[
            [TurnContext, StateT, Any],
            Awaitable[Union[MessagingExtensionResult, TaskModuleTaskInfo, str, None]],
        ]:
            async def __invoke__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False

                value = cast(
                    MessagingExtensionAction,
                    deserializer_helper(MessagingExtensionAction, context.activity.value),
                )
                res = await func(context, state, value.data)
                await self._invoke_action_response(context, res)
                return True

            self._routes.append(Route[StateT](__selector__, __invoke__, True))
            return func

        return __call__

    def _activity_with_command_id(
        self, activity: Activity, match: Union[str, Pattern[str]]
    ) -> bool:
        value = activity.value if isinstance(activity.value, dict) else vars(activity.value)

        if not "commandId" in value and not "command_id" in value:
            return False

        command_id = value["commandId"] if "commandId" in value else value["command_id"]

        if not isinstance(command_id, str):
            return False

        if isinstance(match, Pattern):
            hits = re.match(match, command_id)
            return hits is not None

        return command_id == match

    async def _invoke_response(self, context: TurnContext, body: MessagingExtensionResult):
        if context._INVOKE_RESPONSE_KEY in context.turn_state:
            return

        response = MessagingExtensionResponse(compose_extension=body)

        await context.send_activity(
            Activity(
                type=ActivityTypes.invoke_response,
                value=InvokeResponse(body=serializer_helper(response), status=200),
            )
        )

    async def _invoke_action_response(
        self,
        context: TurnContext,
        body: Union[MessagingExtensionResult, TaskModuleTaskInfo, str, None] = None,
    ):
        if context._INVOKE_RESPONSE_KEY in context.turn_state:
            return

        response = MessagingExtensionActionResponse()

        if isinstance(body, str):
            response = MessagingExtensionActionResponse(
                task=TaskModuleMessageResponse(
                    value=body,
                )
            )
        elif isinstance(body, TaskModuleTaskInfo):
            response = MessagingExtensionActionResponse(task=TaskModuleContinueResponse(value=body))
        elif isinstance(body, MessagingExtensionResult):
            response = MessagingExtensionActionResponse(compose_extension=body)

        await context.send_activity(
            Activity(
                type=ActivityTypes.invoke_response,
                value=InvokeResponse(body=serializer_helper(response), status=200),
            )
        )

    async def _invoke_task_response(
        self, context: TurnContext, body: Union[TaskModuleTaskInfo, str]
    ):
        if context._INVOKE_RESPONSE_KEY in context.turn_state:
            return

        response = TaskModuleResponse(task=TaskModuleContinueResponse(value=body))

        if isinstance(body, str):
            response = TaskModuleResponse(task=TaskModuleMessageResponse(value=body))

        await context.send_activity(
            Activity(
                type=ActivityTypes.invoke_response,
                value=InvokeResponse(body=serializer_helper(response), status=200),
            )
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .message_extensions import MessageExtensions, MessagePreviewAction

__all__ = ["MessageExtensions", "MessagePreviewAction"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from dataclasses import dataclass
from typing import Literal, Optional

SignInStatus = Literal["pending", "complete", "error"]
AuthErrorReason = Literal["completion-without-token", "invalid-activity", "other"]


@dataclass
class SignInResponse:
    status: SignInStatus
    reason: Optional[AuthErrorReason] = None
    message: Optional[str] = None



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import unittest

from teams.ai.tokenizers import GPTTokenizer


class TestGPTTokenizer(unittest.TestCase):
    def setUp(self):
        self.tokenizer = GPTTokenizer()

    def test_encode(self):
        text = "tiktoken is great!"
        expected_result = [83, 1609, 5963, 374, 2294, 0]
        result = self.tokenizer.encode(text)
        self.assertEqual(
            result, expected_result, "Expected result does not match the encoded result"
        )

    def test_decode(self):
        tokens = [83, 1609, 5963, 374, 2294, 0]
        expected_result = "tiktoken is great!"
        result = self.tokenizer.decode(tokens)
        self.assertEqual(
            result, expected_result, "Expected result does not match the decoded result"
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import Any, cast
from unittest import IsolatedAsyncioTestCase, mock

from azure.core.exceptions import HttpResponseError
from botbuilder.core import TurnContext

from teams import ApplicationError
from teams.ai.actions import ActionTypes
from teams.ai.moderators import (
    AzureContentSafetyModerator,
    AzureContentSafetyModeratorOptions,
)
from teams.ai.planners import Plan, PredictedDoCommand, PredictedSayCommand
from teams.ai.prompts.message import Message
from teams.state import ConversationState, TempState, TurnState, UserState


class MockContentSafetyClient:
    def analyze_text(self, *_args, **_kwargs: Any):
        return {}


class MockContentSafetyClientWithResults:
    def analyze_text(self, *_args, **_kwargs: Any):
        return {
            "blocklistsMatchResults": [],
            "hateResult": {"category": "Hate", "severity": 6},
            "selfHarmResult": {"category": "SelfHarm", "severity": 0},
            "sexualResult": {"category": "Sexual", "severity": 0},
            "violenceResult": {"category": "Violence", "severity": 0},
        }


class MockContentSafetyClientWithError:
    def analyze_text(self, *_args, **_kwargs: Any):
        raise HttpResponseError("test")


class TestAzureContentSafetyModerator(IsolatedAsyncioTestCase):
    def create_mock_context(
        self, channel_id="channel1", bot_id="bot1", conversation_id="conversation1", user_id="user1"
    ):
        context = mock.MagicMock()
        context.activity.channel_id = channel_id
        context.activity.recipient.id = bot_id
        context.activity.conversation.id = conversation_id
        context.activity.from_property.id = user_id
        return context

    def test_should_raise_error_no_endpoint(self):
        with self.assertRaises(ApplicationError):
            AzureContentSafetyModerator(
                options=AzureContentSafetyModeratorOptions(api_key="", moderate="output")
            )

    @mock.patch(
        "azure.ai.contentsafety.ContentSafetyClient", return_value=MockContentSafetyClient()
    )
    async def test_should_not_review_input(self, mock_async_openai):
        moderator = AzureContentSafetyModerator(
            options=AzureContentSafetyModeratorOptions(api_key="", moderate="output", endpoint=""),
        )
        plan = await moderator.review_input(context=cast(TurnContext, {}), state=TurnState())
        self.assertTrue(mock_async_openai.called)
        self.assertIsNone(plan)

    @mock.patch(
        "azure.ai.contentsafety.ContentSafetyClient",
        return_value=MockContentSafetyClientWithResults(),
    )
    async def test_should_review_input_and_flag(self, mock_async_openai):
        moderator = AzureContentSafetyModerator(
            options=AzureContentSafetyModeratorOptions(api_key="", moderate="input", endpoint="")
        )
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)
        plan = await moderator.review_input(context=context, state=state)
        self.assertTrue(mock_async_openai.called)
        assert plan is not None
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "DO")
        assert isinstance(plan.commands[0], PredictedDoCommand)
        self.assertEqual(plan.commands[0].action, ActionTypes.FLAGGED_INPUT)

    @mock.patch(
        "azure.ai.contentsafety.ContentSafetyClient",
        return_value=MockContentSafetyClientWithError(),
    )
    async def test_should_review_input_and_error(self, mock_async_openai):
        moderator = AzureContentSafetyModerator(
            options=AzureContentSafetyModeratorOptions(api_key="", moderate="both", endpoint="")
        )
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)
        plan = await moderator.review_input(context=context, state=state)
        self.assertTrue(mock_async_openai.called)
        assert plan is not None
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "DO")
        assert isinstance(plan.commands[0], PredictedDoCommand)
        self.assertEqual(plan.commands[0].action, ActionTypes.HTTP_ERROR)

    @mock.patch(
        "azure.ai.contentsafety.ContentSafetyClient", return_value=MockContentSafetyClient()
    )
    async def test_should_not_review_output(self, mock_async_openai):
        moderator = AzureContentSafetyModerator(
            options=AzureContentSafetyModeratorOptions(api_key="", moderate="input", endpoint="")
        )
        plan = Plan(
            commands=[PredictedSayCommand(response=Message[str](role="assistant", content="test"))]
        )
        output = await moderator.review_output(
            context=cast(TurnContext, {}), state=TurnState(), plan=plan
        )
        self.assertTrue(mock_async_openai.called)
        self.assertEqual(plan, output)

    @mock.patch(
        "azure.ai.contentsafety.ContentSafetyClient",
        return_value=MockContentSafetyClientWithResults(),
    )
    async def test_should_review_output_and_flag(self, mock_async_openai):
        moderator = AzureContentSafetyModerator(
            options=AzureContentSafetyModeratorOptions(api_key="", moderate="output", endpoint="")
        )
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)
        plan = Plan(
            commands=[PredictedSayCommand(response=Message[str](role="assistant", content="test"))]
        )
        output = await moderator.review_output(context=context, state=state, plan=plan)
        self.assertTrue(mock_async_openai.called)
        assert output is not None
        self.assertEqual(len(output.commands), 1)
        self.assertEqual(output.commands[0].type, "DO")
        assert isinstance(output.commands[0], PredictedDoCommand)
        self.assertEqual(output.commands[0].action, ActionTypes.FLAGGED_OUTPUT)

    @mock.patch(
        "azure.ai.contentsafety.ContentSafetyClient",
        return_value=MockContentSafetyClientWithError(),
    )
    async def test_should_review_output_and_error(self, mock_async_openai):
        moderator = AzureContentSafetyModerator(
            options=AzureContentSafetyModeratorOptions(api_key="", moderate="both", endpoint="")
        )
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)
        plan = Plan(
            commands=[PredictedSayCommand(response=Message[str](role="assistant", content="test"))]
        )
        output = await moderator.review_output(context=context, state=state, plan=plan)
        self.assertTrue(mock_async_openai.called)
        assert output is not None
        self.assertEqual(len(output.commands), 1)
        self.assertEqual(output.commands[0].type, "DO")
        assert isinstance(output.commands[0], PredictedDoCommand)
        self.assertEqual(output.commands[0].action, ActionTypes.HTTP_ERROR)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import TestCase

from teams.ai.prompts import UserMessage


class TestUserMessage(TestCase):
    def test_init(self):
        user_message = UserMessage("template")
        self.assertEqual(user_message.template, "template")
        self.assertEqual(user_message.tokens, -1)
        self.assertEqual(user_message.text_prefix, "user: ")
        self.assertEqual(user_message.role, "user")
        self.assertTrue(user_message.required)
        self.assertEqual(user_message.separator, "\n")

    def test_init_with_optional_params(self):
        user_message = UserMessage("template", 100, "text_prefix")
        self.assertEqual(user_message.template, "template")
        self.assertEqual(user_message.tokens, 100)
        self.assertEqual(user_message.text_prefix, "text_prefix")
        self.assertEqual(user_message.role, "user")
        self.assertTrue(user_message.required)
        self.assertEqual(user_message.separator, "\n")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import List, Optional, TypeVar

from botbuilder.core import TurnContext
from botbuilder.schema import Activity, TokenResponse
from botframework.connector.token_api.models import TokenExchangeRequest

from ...state import TurnState
from ..auth import Auth
from ..auth_component import AuthComponent
from .oauth_adaptive_card import OAuthAdaptiveCard
from .oauth_dialog import OAuthDialog
from .oauth_message_extension import OAuthMessageExtension
from .oauth_options import OAuthOptions

StateT = TypeVar("StateT", bound=TurnState)


class OAuth(Auth[StateT]):
    "Handles authentication using OAuth Connection."

    _options: OAuthOptions
    _components: List[AuthComponent[StateT]]

    def __init__(self, options: OAuthOptions) -> None:
        super().__init__()
        self._options = options
        self._components = [
            OAuthDialog[StateT](options),
            OAuthMessageExtension[StateT](options),
            OAuthAdaptiveCard[StateT](options),
        ]

    def is_sign_in_activity(self, activity: Activity) -> bool:
        for component in self._components:
            if component.is_sign_in_activity(activity):
                return True

        return False

    async def get_token(self, context: TurnContext) -> Optional[str]:
        client = self._user_token_client(context)
        res = await client.get_user_token(
            getattr(context.activity.from_property, "id"),
            self._options.connection_name,
            context.activity.channel_id,
            "",
        )

        if res and res.token:
            return res.token

        return None

    async def sign_in(self, context: TurnContext, state: StateT) -> Optional[str]:
        token = await self.get_token(context)

        if token:
            return token

        for component in self._components:
            if component.is_sign_in_activity(context.activity):
                state.user["__auth__"] = state.temp.input
                return await component.sign_in(context, state)

        return None

    async def sign_out(self, context: TurnContext, state: StateT) -> None:
        client = self._user_token_client(context)

        for component in self._components:
            await component.sign_out(context, state)

        await client.sign_out_user(
            getattr(context.activity.from_property, "id"),
            self._options.connection_name,
            context.activity.channel_id,
        )

    async def verify_state(self, context: TurnContext, state: StateT) -> Optional[TokenResponse]:
        client = self._user_token_client(context)
        value = context.activity.value
        code: Optional[str] = None

        if value is None:
            return None

        if hasattr(value, "state") and isinstance(value.state, str):
            code = getattr(value, "state")

        if isinstance(value, dict) and "state" in value:
            code = value["state"]

        if code is None:
            return None

        if "__auth__" in state.user:
            state.temp.input = state.user["__auth__"]
            del state.user["__auth__"]

        return await client.get_user_token(
            getattr(context.activity.from_property, "id"),
            self._options.connection_name,
            context.activity.channel_id,
            code,
        )

    async def exchange_token(self, context: TurnContext, state: StateT) -> Optional[TokenResponse]:
        client = self._user_token_client(context)

        if not context.activity.value:
            return None

        if "__auth__" in state.user:
            state.temp.input = state.user["__auth__"]
            del state.user["__auth__"]

        return await client.exchange_token(
            getattr(context.activity.from_property, "id"),
            self._options.connection_name,
            context.activity.channel_id,
            TokenExchangeRequest(token=context.activity.value.get("token")),
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Optional

from botbuilder.core import TurnContext

from ...state import TurnState
from ..planners.plan import Plan
from .moderator import Moderator


class DefaultModerator(Moderator):
    """
    Default moderator created by the AI system if one isn't configured.
    """

    async def review_input(self, context: TurnContext, state: TurnState) -> Optional[Plan]:
        return None

    async def review_output(self, context: TurnContext, state: TurnState, plan: Plan) -> Plan:
        return plan



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase

from botbuilder.core import TurnContext

from teams.ai.augmentations.default_augmentation import DefaultAugmentation
from teams.ai.models.prompt_response import PromptResponse
from teams.ai.planners.plan import Plan, PredictedDoCommand, PredictedSayCommand
from teams.ai.prompts.message import Message
from teams.ai.tokenizers.gpt_tokenizer import GPTTokenizer
from teams.state import TurnState


class TestDefaultAugmentation(IsolatedAsyncioTestCase):
    def setUp(self):
        self.tokenizer = GPTTokenizer()
        self.default_augmentation = DefaultAugmentation()

    def test_create_prompt_section(self):
        self.assertEqual(self.default_augmentation.create_prompt_section(), None)

    async def test_validate_response(self):
        state = TurnState()
        prompt_response = PromptResponse(
            message=Message(
                role="assistant",
                content=str(
                    Plan(commands=[PredictedDoCommand(action="test", parameters={"foo": "bar"})])
                ),
            )
        )
        validation = await self.default_augmentation.validate_response(
            cast(TurnContext, {}),
            memory=state,
            tokenizer=self.tokenizer,
            response=prompt_response,
            remaining_attempts=0,
        )
        self.assertEqual(validation.valid, True)

    async def test_create_plan_from_response_where_content_exists(self):
        state = TurnState()

        # Validate response
        say_response = PromptResponse(
            message=Message(
                role="assistant",
                content='{ "type": "plan", '
                + '"commands": [{ "type": "SAY", "response": { "role": "assistant", "content":'
                ' "hello world"}}]}',
            )
        )
        validation = await self.default_augmentation.validate_response(
            cast(TurnContext, {}),
            memory=state,
            tokenizer=self.tokenizer,
            response=say_response,
            remaining_attempts=3,
        )
        self.assertEqual(validation.valid, True)

        # Create plan from validated response
        plan = await self.default_augmentation.create_plan_from_response(
            cast(TurnContext, {}),
            memory=state,
            response=PromptResponse[str](
                message=Message(role="assistant", content=validation.value or "")
            ),
        )
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "SAY")
        assert isinstance(plan.commands[0], PredictedSayCommand)
        self.assertEqual(
            plan.commands[0].response,
            Message(role="assistant", content="", function_call=None, name=None),
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .adaptive_cards_options import AdaptiveCardsOptions
from .adaptive_cards_search_params import AdaptiveCardsSearchParams
from .adaptive_cards_search_result import AdaptiveCardsSearchResult

__all__ = ["AdaptiveCardsOptions", "AdaptiveCardsSearchParams", "AdaptiveCardsSearchResult"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase

from botbuilder.core import TurnContext

from teams.ai.models import PromptResponse
from teams.ai.tokenizers import GPTTokenizer
from teams.ai.validators import DefaultResponseValidator
from teams.state import TurnState


class TestDefaultResponseValidator(IsolatedAsyncioTestCase):
    async def test_should_be_valid(self):
        validator = DefaultResponseValidator()
        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=PromptResponse(),
            remaining_attempts=3,
        )

        self.assertTrue(res.valid)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from botbuilder.core import TurnContext

from ...state import TurnState
from .plan import Plan

StateT = TypeVar("StateT", bound=TurnState)


class Planner(Generic[StateT], ABC):
    """
    A planner is responsible for generating a plan that the AI system will execute.
    """

    @abstractmethod
    async def begin_task(self, context: TurnContext, state: TurnState) -> Plan:
        """
        Starts a new task.

        Args:
            context (TurnContext): The current turn context.
            state (TurnState): The current turn state.
        """

    @abstractmethod
    async def continue_task(self, context: TurnContext, state: TurnState) -> Plan:
        """
        Continues the current task.

        Args:
            context (TurnContext): The current turn context.
            state (TurnState): The current turn state.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import List

from .sections.layout_engine_section import LayoutEngineSection
from .sections.prompt_section import PromptSection


class Prompt(LayoutEngineSection):
    """
    Top level prompt section.

    Prompts are compositional such that they can be nested to create complex prompt hierarchies.
    """

    def __init__(
        self,
        sections: List[PromptSection],
        tokens: float = -1,
        required: bool = True,
        separator: str = "\n\n",
    ):
        """
        Creates a new 'Prompt' instance.

        Args:
            sections (List[PromptSection]): Sections to render.
            tokens (float, optional): Sizing strategy for this section. Defaults to `auto`.
            required (bool, optional): Indicates if this section is required. Defaults to `true`.
            separator (str, optional): Separator to use between sections when rendering as text.
              Defaults to `\n\n`.
        """
        super().__init__(sections, tokens, required, separator)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import List, Optional

from botbuilder.core import TurnContext

from teams.ai.data_sources.data_source import DataSource
from teams.ai.prompts.rendered_prompt_section import RenderedPromptSection
from teams.ai.tokenizers import Tokenizer
from teams.state.memory import MemoryBase


class TextDataSource(DataSource):
    """
    A data source that can be used to add a static block of text to a prompt.

    .. remarks::
    Primarily used for testing but could be used to inject some externally define text
    into a prompt. The text will be truncated to fit within the `max_tokens` limit.
    """

    _name: str
    _text: str
    _tokens: Optional[List[int]] = None

    def __init__(self, name: str, text: str) -> None:
        """
        Creates a new `TextDataSource` instance.

        Args:
            name (str): Name of the data source.
            text (str): Text to inject into the prompt.
        """
        self._name = name
        self._text = text

    @property
    def name(self) -> str:
        """
        Name of the data source.
        """
        return self._name

    async def render_data(
        self,
        turn_context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[str]:
        """
        Renders the data source as a string of text.

        Args:
            turn_context (TurnContext): The turn context for current turn of conversation.
            memory (MemoryBase): An interface for accessing state values.
            tokenizer (Tokenizer): Tokenizer to use when rendering the data source.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection: The text to inject into the prompt as a
            `RenderedPromptSection` object.
        """

        if not self._tokens:  # Tokenize text on first use
            self._tokens = tokenizer.encode(self._text)

        if len(self._tokens) > max_tokens:  # Check for max tokens
            trimmed = self._tokens[0:max_tokens]
            return RenderedPromptSection[str](
                output=tokenizer.decode(trimmed), length=len(trimmed), too_long=True
            )
        return RenderedPromptSection[str](
            output=self._text, length=len(self._tokens), too_long=False
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import TestCase

from teams.ai.citations.citations import Appearance, ClientCitation
from teams.utils.citations import get_used_citations


class TestGetUsedCitations(TestCase):
    def test_get_used_citations(self):
        citations = [
            ClientCitation(
                position="1",
                appearance=Appearance(name="the title", abstract="some citation text..."),
            ),
            ClientCitation(
                position="2",
                appearance=Appearance(name="the title", abstract="some citation other text..."),
            ),
        ]
        text = "hello [1] world"
        result = get_used_citations(text, citations)
        assert result is not None
        self.assertEqual(len(result), 1)

        self.assertEqual(result, [citations[0]])

    def test_get_used_citations_longer(self):
        citations = [
            ClientCitation(
                position="1",
                appearance=Appearance(name="the title", abstract="some citation text..."),
            ),
            ClientCitation(
                position="2",
                appearance=Appearance(name="the title", abstract="some citation other text..."),
            ),
            ClientCitation(
                position="3",
                appearance=Appearance(name="the title", abstract="some citation text..."),
            ),
            ClientCitation(
                position="4",
                appearance=Appearance(name="the title", abstract="some citation other text..."),
            ),
        ]
        text = "hello [1] world [3]"
        result = get_used_citations(text, citations)
        if result is not None:
            self.assertEqual(len(result), 2)
        else:
            self.fail("Result is None")
        self.assertEqual(result, [citations[0], citations[2]])



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import List, Literal, Optional, Union, cast
from unittest import IsolatedAsyncioTestCase, mock

import httpx
import openai
from botbuilder.core import TurnContext

from teams.ai.actions import ActionTypes
from teams.ai.moderators import OpenAIModerator, OpenAIModeratorOptions
from teams.ai.planners import Plan, PredictedDoCommand, PredictedSayCommand
from teams.ai.prompts.message import Message
from teams.state import ConversationState, TempState, TurnState, UserState


class MockAsyncModerations:
    async def create(
        self,
        *,
        input: Union[str, List[str]],
        model: Union[
            str, Literal["text-moderation-latest", "text-moderation-stable"]
        ] = "text-moderation-latest",
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> openai.types.ModerationCreateResponse:
        # pylint: disable=unused-argument
        return openai.types.ModerationCreateResponse(id="", model=model, results=[])


class MockAsyncModerationsWithResults:
    async def create(
        self,
        *,
        input: Union[str, List[str]],
        model: Union[
            str, Literal["text-moderation-latest", "text-moderation-stable"]
        ] = "text-moderation-latest",
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> openai.types.ModerationCreateResponse:
        # pylint: disable=unused-argument
        return openai.types.ModerationCreateResponse(
            id="",
            model=model,
            results=[
                openai.types.Moderation(
                    categories=cast(
                        openai.types.moderation.Categories,
                        {
                            "harassment": True,
                            "harassment/threatening": False,
                            "hate": False,
                            "hate/threatening": False,
                            "self-harm": False,
                            "self-harm/instructions": False,
                            "self-harm/intent": False,
                            "sexual": False,
                            "sexual/minors": False,
                            "violence": False,
                            "violence/graphic": False,
                        },
                    ),
                    category_scores=cast(
                        openai.types.moderation.CategoryScores,
                        {
                            "harassment": 0,
                            "harassment/threatening": 0,
                            "hate": 0,
                            "hate/threatening": 0,
                            "self-harm": 0,
                            "self-harm/instructions": 0,
                            "self-harm/intent": 0,
                            "sexual": 0,
                            "sexual/minors": 0,
                            "violence": 0,
                            "violence/graphic": 0,
                        },
                    ),
                    flagged=True,
                )
            ],
        )


class MockAsyncModerationsRateLimited:
    async def create(
        self,
        *,
        input: Union[str, List[str]],
        model: Union[
            str, Literal["text-moderation-latest", "text-moderation-stable"]
        ] = "text-moderation-latest",
        extra_headers: Optional[openai._types.Headers] = None,
        extra_query: Optional[openai._types.Query] = None,
        extra_body: Optional[openai._types.Body] = None,
        timeout: Union[
            float, httpx.Timeout, None, openai._types.NotGiven
        ] = openai._types.NOT_GIVEN,
    ) -> openai.types.ModerationCreateResponse:
        # pylint: disable=unused-argument
        raise openai.RateLimitError(
            message="This is a rate limited error",
            response=httpx.Response(
                status_code=429, request=httpx.Request(method="method", url="url")
            ),
            body=None,
        )


class MockAsyncOpenAI:
    moderations = MockAsyncModerations()


class MockAsyncOpenAIWithResults:
    moderations = MockAsyncModerationsWithResults()


class MockAsyncOpenAIRateLimited:
    moderations = MockAsyncModerationsRateLimited()


class TestOpenAIModerator(IsolatedAsyncioTestCase):
    def create_mock_context(
        self, channel_id="channel1", bot_id="bot1", conversation_id="conversation1", user_id="user1"
    ):
        context = mock.MagicMock()
        context.activity.channel_id = channel_id
        context.activity.recipient.id = bot_id
        context.activity.conversation.id = conversation_id
        context.activity.from_property.id = user_id
        return context

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAI)
    async def test_should_not_review_input(self, mock_async_openai):
        moderator = OpenAIModerator(options=OpenAIModeratorOptions(api_key="", moderate="output"))
        plan = await moderator.review_input(context=cast(TurnContext, {}), state=TurnState())
        self.assertTrue(mock_async_openai.called)
        self.assertIsNone(plan)

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAIWithResults)
    async def test_should_review_input_and_flag(self, mock_async_openai):
        moderator = OpenAIModerator(options=OpenAIModeratorOptions(api_key="", moderate="input"))
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)
        plan = await moderator.review_input(context=context, state=state)
        self.assertTrue(mock_async_openai.called)
        assert plan is not None
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "DO")
        assert isinstance(plan.commands[0], PredictedDoCommand)
        self.assertEqual(plan.commands[0].action, ActionTypes.FLAGGED_INPUT)

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAIRateLimited)
    async def test_should_review_input_and_error(self, mock_async_openai):
        moderator = OpenAIModerator(options=OpenAIModeratorOptions(api_key="", moderate="both"))
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)
        plan = await moderator.review_input(context=context, state=state)
        self.assertTrue(mock_async_openai.called)
        assert plan is not None
        self.assertEqual(len(plan.commands), 1)
        self.assertEqual(plan.commands[0].type, "DO")
        assert isinstance(plan.commands[0], PredictedDoCommand)
        self.assertEqual(plan.commands[0].action, ActionTypes.HTTP_ERROR)

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAI)
    async def test_should_not_review_output(self, mock_async_openai):
        moderator = OpenAIModerator(options=OpenAIModeratorOptions(api_key="", moderate="input"))
        plan = Plan(
            commands=[PredictedSayCommand(response=Message[str](role="assistant", content="test"))]
        )
        output = await moderator.review_output(
            context=cast(TurnContext, {}), state=TurnState(), plan=plan
        )
        self.assertTrue(mock_async_openai.called)
        self.assertEqual(plan, output)

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAIWithResults)
    async def test_should_review_output_and_flag(self, mock_async_openai):
        moderator = OpenAIModerator(options=OpenAIModeratorOptions(api_key="", moderate="output"))
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)
        plan = Plan(
            commands=[PredictedSayCommand(response=Message[str](role="assistant", content="test"))]
        )
        output = await moderator.review_output(context=context, state=state, plan=plan)
        self.assertTrue(mock_async_openai.called)
        assert output is not None
        self.assertEqual(len(output.commands), 1)
        self.assertEqual(output.commands[0].type, "DO")
        assert isinstance(output.commands[0], PredictedDoCommand)
        self.assertEqual(output.commands[0].action, ActionTypes.FLAGGED_OUTPUT)

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAIRateLimited)
    async def test_should_review_output_and_error(self, mock_async_openai):
        moderator = OpenAIModerator(options=OpenAIModeratorOptions(api_key="", moderate="both"))
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)
        plan = Plan(
            commands=[PredictedSayCommand(response=Message[str](role="assistant", content="test"))]
        )
        output = await moderator.review_output(context=context, state=state, plan=plan)
        self.assertTrue(mock_async_openai.called)
        assert output is not None
        self.assertEqual(len(output.commands), 1)
        self.assertEqual(output.commands[0].type, "DO")
        assert isinstance(output.commands[0], PredictedDoCommand)
        self.assertEqual(output.commands[0].action, ActionTypes.HTTP_ERROR)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import re
from typing import Awaitable, Callable, Generic, List, Pattern, TypeVar, Union

from botbuilder.core import TurnContext
from botbuilder.schema import (
    Activity,
    ActivityTypes,
    AdaptiveCardInvokeResponse,
    InvokeResponse,
)

from teams.query import Query
from teams.route import Route
from teams.state import TurnState

from .adaptive_cards_search_params import AdaptiveCardsSearchParams
from .adaptive_cards_search_result import AdaptiveCardsSearchResult

ACTION_INVOKE_NAME = "adaptiveCard/action"
ACTION_EXECUTE_TYPE = "Action.Execute"
SEARCH_INVOKE_NAME = "application/search"

StateT = TypeVar("StateT", bound=TurnState)


class AdaptiveCards(Generic[StateT]):
    _route_registry: List[Route]
    _action_submit_filter: str

    def __init__(self, route_registry: List[Route], action_submit_filter: str) -> None:
        self._route_registry = route_registry
        self._action_submit_filter = action_submit_filter

    def action_execute(
        self, verb: Union[str, Pattern[str], Callable[[TurnContext], bool]]
    ) -> Callable[
        [Callable[[TurnContext, StateT, dict], Awaitable[Union[str, dict]]]],
        Callable[[TurnContext, StateT, dict], Awaitable[Union[str, dict]]],
    ]:
        """
        Adds a route for handling Adaptive Card Action.Execute events.
        This method can be used as either a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.adaptive_cards.action_execute("do")
        async def execute_do(context: TurnContext, state: TurnState, data: Any):
            print(f"Execute with data: {data}")
            return True

        # Pass a function to this method
        app.adaptive_cards.action_execute("do")(execute_do)
        ```

        #### Args:
        - `verb`: a string, regex pattern or a function to match the verb of the action
        """

        # Create route selector for the handler
        def __selector__(context: TurnContext) -> bool:
            if (
                context.activity.type == ActivityTypes.invoke
                and context.activity.name == ACTION_INVOKE_NAME
            ):
                action = context.activity.value["action"] if context.activity.value else None
                action_type = action["type"] if action else None
                action_verb = action["verb"] if action else None
                if action_type == ACTION_EXECUTE_TYPE:
                    # when verb is a function
                    if callable(verb):
                        return verb(context)
                    # when verb is a regex pattern
                    if isinstance(verb, Pattern) and isinstance(action_verb, str):
                        hits = re.match(verb, action_verb)
                        return hits is not None
                    # when verb is a string
                    return verb == action_verb

            return False

        def __call__(
            func: Callable[[TurnContext, StateT, dict], Awaitable[Union[str, dict]]],
        ) -> Callable[[TurnContext, StateT, dict], Awaitable[Union[str, dict]]]:
            async def __handler__(context: TurnContext, state: StateT) -> bool:
                result = await func(context, state, context.activity.value["action"]["data"])

                if context.turn_state.get(context._INVOKE_RESPONSE_KEY) is None:
                    if isinstance(result, str):
                        response = AdaptiveCardInvokeResponse(
                            status_code=200,
                            type="application/vnd.microsoft.activity.message",
                            value=result,
                        )
                    else:
                        response = AdaptiveCardInvokeResponse(
                            status_code=200,
                            type="application/vnd.microsoft.card.adaptive",
                            value=result,
                        )
                    await context.send_activity(
                        Activity(
                            type="invokeResponse", value=InvokeResponse(status=200, body=response)
                        )
                    )
                return True

            self._route_registry.append(Route[StateT](__selector__, __handler__))
            return func

        return __call__

    def action_submit(
        self, verb: Union[str, Pattern[str], Callable[[TurnContext], bool]]
    ) -> Callable[
        [Callable[[TurnContext, StateT, dict], Awaitable[None]]],
        Callable[[TurnContext, StateT, dict], Awaitable[None]],
    ]:
        """
        Adds a route for handling Adaptive Card Action.Submit events.
        This method can be used as either a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.adaptive_cards.action_submit("submit")
        async def execute_submit(context: TurnContext, state: TurnState, data: Any):
            print(f"Execute with data: {data}")
            return True

        # Pass a function to this method
        app.adaptive_cards.action_submit("submit")(execute_submit)
        ```

        #### Args:
        - `verb`: a string, regex pattern or a function to match the verb of the action
        """

        # Create route selector for the handler
        def __selector__(context: TurnContext) -> bool:
            filter = self._action_submit_filter
            if (
                context.activity.type == ActivityTypes.message
                and context.activity.text is None
                and isinstance(context.activity.value, dict)
            ):
                filter_value = context.activity.value.get(filter)
                # when verb is a function
                if callable(verb):
                    return verb(context)
                # when verb is a regex pattern
                if isinstance(verb, Pattern) and isinstance(filter_value, str):
                    hits = re.match(verb, filter_value)
                    return hits is not None
                # when verb is a string
                return verb == filter_value

            return False

        def __call__(
            func: Callable[[TurnContext, StateT, dict], Awaitable[None]],
        ) -> Callable[[TurnContext, StateT, dict], Awaitable[None]]:
            async def __handler__(context: TurnContext, state: StateT) -> bool:
                await func(context, state, context.activity.value)
                return True

            self._route_registry.append(Route[StateT](__selector__, __handler__))
            return func

        return __call__

    def search(self, dataset: Union[str, Pattern[str], Callable[[TurnContext], bool]]) -> Callable[
        [
            Callable[
                [TurnContext, StateT, Query[AdaptiveCardsSearchParams]],
                Awaitable[List[AdaptiveCardsSearchResult]],
            ]
        ],
        Callable[
            [TurnContext, StateT, Query[AdaptiveCardsSearchParams]],
            Awaitable[List[AdaptiveCardsSearchResult]],
        ],
    ]:
        """
        Adds a route for handling Adaptive Card search events. This method can be used as either
        a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.adaptive_cards.search("people")
        async def search_people(context: TurnContext, state: TurnState, query: Any):
            print(f"Execute with query: {query}")
            return True

        # Pass a function to this method
        app.adaptive_cards.search("people")(search_people)
        ```

        #### Args:
        - `dataset`: a string, regex pattern or a function to match the verb of the action
        """

        # Create route selector for the handler
        def __selector__(context: TurnContext) -> bool:
            if (
                context.activity.type == ActivityTypes.invoke
                and context.activity.name == SEARCH_INVOKE_NAME
            ):
                activity_dataset = (
                    context.activity.value["dataset"] if context.activity.value else None
                )
                # when verb is a function
                if callable(dataset):
                    return dataset(context)
                # when verb is a regex pattern
                if isinstance(dataset, Pattern) and isinstance(activity_dataset, str):
                    hits = re.match(dataset, activity_dataset)
                    return hits is not None
                # when verb is a string
                return dataset == activity_dataset

            return False

        def __call__(
            func: Callable[
                [TurnContext, StateT, Query[AdaptiveCardsSearchParams]],
                Awaitable[List[AdaptiveCardsSearchResult]],
            ],
        ) -> Callable[
            [TurnContext, StateT, Query[AdaptiveCardsSearchParams]],
            Awaitable[List[AdaptiveCardsSearchResult]],
        ]:
            async def __handler__(context: TurnContext, state: StateT) -> bool:
                params = context.activity.value
                # Flatten search parameters
                query = Query[AdaptiveCardsSearchParams](
                    count=(
                        params["queryOptions"]["top"]
                        if params and params["queryOptions"] and params["queryOptions"]["top"]
                        else 25
                    ),
                    skip=(
                        params["queryOptions"]["skip"]
                        if params and params["queryOptions"] and params["queryOptions"]["skip"]
                        else 0
                    ),
                    parameters=AdaptiveCardsSearchParams(
                        query_text=params["queryText"] if params and params["queryText"] else "",
                        dataset=params["dataset"] if params and params["dataset"] else "",
                    ),
                )
                result = await func(context, state, query)
                if context.turn_state.get(context._INVOKE_RESPONSE_KEY) is None:
                    # Format invoke response
                    response = {
                        "type": "application/vnd.microsoft.search.searchResponse",
                        "value": {"results": result},
                    }
                    await context.send_activity(
                        Activity(
                            type="invokeResponse", value=InvokeResponse(status=200, body=response)
                        )
                    )
                return True

            self._route_registry.append(Route[StateT](__selector__, __handler__))
            return func

        return __call__



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, Literal, Optional, TypeVar, Union

from dataclasses_json import DataClassJsonMixin, dataclass_json

from .function_call import FunctionCall

T = TypeVar("T")


@dataclass_json
@dataclass
class Message(Generic[T], DataClassJsonMixin):
    """
    A message object sent to or received from an LLM.

    Attributes:
        role (str): The messages role. Typically 'system', 'user', 'assistant', 'function'.
        content (Optional[T]): Text of the message.
        function_call (Optional[FunctionCall]): A named function to call.
        name (Optional[str]): Name of the function that was called.
    """

    role: str
    content: Optional[T] = None
    context: Optional[MessageContext] = None
    function_call: Optional[FunctionCall] = None
    name: Optional[str] = None


@dataclass
class ImageUrl:
    """
    Url for an image

    Attributes:
        url (str): Url of the image
    """

    url: str


@dataclass
class TextContentPart:
    """
    Represents text content part of a message

    Attributes:
        text (str): Text content
    """

    type: Literal["text"]
    text: str


@dataclass
class ImageContentPart:
    """
    Represents image content part of a message

    Attributes:
        image_url (Union[str, ImageUrl]): Url for the image
    """

    type: Literal["image_url"]
    image_url: Union[str, ImageUrl]


MessageContentParts = Union[TextContentPart, ImageContentPart]
"""
Represents part of the message's content
"""


@dataclass
class Citation:
    """
    Represents a citation returned by the model

    Attributes:
        content (str): The content of the citation
        title (str): The title of the citation
        url (str): The url of the citation
        filepath (str): The filepath of the citation
    """

    content: str
    title: str
    url: str
    filepath: str


@dataclass
class MessageContext:
    """
    Represents the message context containing a citation

    Attributes:
        citations (list[Citation]): The citations in the message
        intent (str): The intent of the message
    """

    citations: list[Citation]
    intent: str



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .augmentation import Augmentation
from .default_augmentation import DefaultAugmentation
from .monologue_augmentation import MonologueAugmentation
from .sequence_augmentation import SequenceAugmentation

__all__ = [
    "Augmentation",
    "DefaultAugmentation",
    "MonologueAugmentation",
    "SequenceAugmentation",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass

from teams.teams_adapter import TeamsAdapter


@dataclass
class TeamsAttachmentDownloaderOptions:
    """
    Options for the `TeamsAttachmentDownloader` class.
    """

    bot_app_id: str
    "The Microsoft App ID of the bot"

    adapter: TeamsAdapter
    "ServiceClientCredentialsFactory"



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import re


def format_citations_response(text: str) -> str:
    """
    Convert citation tags `[doc(s)n]` to `[n]` where n is a number.
    Args:
        text: str The text to format.
    Returns:
        str: The modified text
    """
    return re.sub(r"\[docs?(\d+)\]", r"[\1]", text, flags=re.IGNORECASE)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, Optional, TypeVar

from .rendered_prompt_section import RenderedPromptSection
from .sections.prompt_section import PromptSection

T = TypeVar("T")


@dataclass
class _PromptSectionLayout(Generic[T]):
    section: PromptSection
    layout: Optional[RenderedPromptSection[T]] = None



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import List

from botbuilder.core import TurnContext

from ....state import MemoryBase
from ...tokenizers import Tokenizer
from ..message import Message
from ..prompt_functions import PromptFunctions
from ..rendered_prompt_section import RenderedPromptSection
from .layout_engine_section import LayoutEngineSection
from .prompt_section import PromptSection
from .prompt_section_base import PromptSectionBase


class GroupSection(PromptSectionBase):
    """
    A group of sections that will rendered as a single message.
    """

    _layout_engine: LayoutEngineSection
    _sections: List[PromptSection]
    _role: str

    def __init__(
        self,
        sections: List[PromptSection],
        role: str = "system",
        tokens: float = -1,
        required: bool = True,
        separator: str = "\n\n",
        text_prefix: str = "",
    ):
        """
        Initializes the GroupSection object.

        Args:
            sections (List[PromptSection]): List of sections to group together.
            role (str, optional): Message role to use for this section. Defaults to `system`.
            tokens (float, optional): Sizing strategy for this section. Defaults to `auto`.
            required (bool, optional): Indicates if this section is required. Defaults to `true`.
            separator (str, optional): Separator to use between sections when rendering as text.
              Defaults to `\n\n`.
            text_prefix (str, optional): Prefix to use for text output. Defaults to ``.
        """
        super().__init__(tokens, required, separator, text_prefix)
        self._layout_engine = LayoutEngineSection(sections, tokens, required, separator)
        self._sections = sections
        self._role = role

    @property
    def sections(self):
        """
        Gets the sections in this group.
        """
        return self._sections

    @property
    def role(self):
        """
        Gets the role for this group.
        """
        return self._role

    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message]]:
        """
        Renders the section as a list of messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered prompt section as a list of messages.
        """
        # Render sections to text
        result = await self._layout_engine.render_as_text(
            context, memory, functions, tokenizer, max_tokens
        )

        # Return output as a single message
        return self._return_messages(
            [Message(self.role, result.output)], result.length, tokenizer, max_tokens
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Union

from teams.ai.embeddings.embeddings_response import EmbeddingsResponse

from ...user_agent import _UserAgent


class EmbeddingsModel(ABC, _UserAgent):
    """
    An AI model that can be used to create embeddings.
    """

    @abstractmethod
    async def create_embeddings(
        self, inputs: Union[str, List[str], List[int], List[List[int]]]
    ) -> EmbeddingsResponse:
        """
        Creates embeddings for the given inputs.

        Args:
            inputs (Union[str, List[str],
            List[int], List[List[int]]]): Text inputs to create embeddings for.

        Returns:
            EmbeddingsResponse: A status and embeddings/message when an error occurs.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import TestCase

from teams.utils import snippet


class TestSnippet(TestCase):
    def test_snippet(self):
        text = "This is a test snippet"
        self.assertEqual(snippet(text, 10), "This is a...")
        self.assertEqual(snippet(text, 100), text)
        hello = "hello world"
        self.assertEqual(snippet(hello, 5), "hell...")
        self.assertEqual(snippet(hello, 6), "hello...")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Dict, Generic, Optional, TypeVar, cast

from botbuilder.core import TurnContext
from botbuilder.schema import Activity, ActivityTypes, InvokeResponse, SignInConstants

from ..state import TempState, TurnState
from .auth import Auth
from .sign_in_response import SignInResponse

StateT = TypeVar("StateT", bound=TurnState)


class AuthManager(Generic[StateT]):
    "authentication manager"

    _default: Optional[str]
    _connections: Dict[str, Auth[StateT]] = {}
    _exchanges: Dict[str, str] = {}

    def __init__(self, *, default: Optional[str] = None) -> None:
        self._default = default

    def get(self, key: str) -> Auth[StateT]:
        """
        Gets the Auth connection

        Args:
            key (str): the auth connection key

        Returns:
            the auth connection if exists, otherwise raises a RuntimeError
        """

        if not key in self._connections:
            raise RuntimeError(f"could not find auth connection '{key}'")

        return self._connections[key]

    def set(self, key: str, value: Auth[StateT]) -> None:
        """
        Sets a new Auth connection

        Args:
            key (str): the auth connection key
            value (Auth[StateT]): the auth connection
        """
        self._connections[key] = value

    async def sign_in(
        self, context: TurnContext, state: StateT, *, key: Optional[str] = None
    ) -> SignInResponse:
        key = key if key is not None else self._default

        if not key:
            raise ValueError("must specify a connection 'key'")

        auth = self.get(key)
        token: Optional[str] = await auth.get_token(context)
        res = SignInResponse("pending")

        if token:
            cast(TempState, state.temp).auth_tokens[key] = token
            return SignInResponse("complete")

        try:
            if self._is_exchange_activity(context):
                exchange_key = self._get_exchange_key(key, context)

                if exchange_key in self._exchanges:
                    return res

                self._exchanges[exchange_key] = cast(dict, context.activity.value)["id"]
                token_res = await auth.exchange_token(context, state)

                if token_res is not None and hasattr(token_res, "token"):
                    token = token_res.token
                else:
                    # tell client to prompt for consent
                    await context.send_activity(
                        Activity(
                            type=ActivityTypes.invoke_response, value=InvokeResponse(status=412)
                        )
                    )

                    return res

                del self._exchanges[exchange_key]
            elif self._is_verify_state_activity(context):
                token_res = await auth.verify_state(context, state)

                if token_res is not None:
                    token = token_res.token
            else:
                token = await auth.sign_in(context, state)
        except Exception as err:  # pylint:disable=broad-exception-caught
            res.status = "error"
            res.reason = "other"
            res.message = str(err)

            if auth._on_sign_in_failure:
                await auth._on_sign_in_failure(context, state, res)

        if token:
            cast(TempState, state.temp).auth_tokens[key] = token
            res.status = "complete"

            if auth._on_sign_in_success:
                await auth._on_sign_in_success(context, state)

        return res

    async def sign_out(
        self, context: TurnContext, state: StateT, *, key: Optional[str] = None
    ) -> None:
        key = key if key is not None else self._default

        if not key:
            raise ValueError("must specify a connection 'key'")

        auth = self.get(key)
        await auth.sign_out(context, state)

    def _is_exchange_activity(self, context: TurnContext) -> bool:
        return context.activity.type == ActivityTypes.invoke and (
            context.activity.name == SignInConstants.token_exchange_operation_name
        )

    def _is_verify_state_activity(self, context: TurnContext) -> bool:
        return context.activity.type == ActivityTypes.invoke and (
            context.activity.name == SignInConstants.verify_state_operation_name
        )

    def _get_exchange_key(self, key: str, context: TurnContext) -> str:
        return (
            cast(str, getattr(context.activity.from_property, "id"))
            + "/"
            + cast(str, getattr(context.activity.conversation, "id"))
            + "/"
            + key
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional


@dataclass
class Validation:
    """
    Response returned by a `PromptResponseValidator`.
    """

    valid: bool = True
    """
    Whether the validation is valid.
    Default: True
    """

    feedback: Optional[str] = None
    """
    Optional. Repair instructions to send to the model.
    Default: None
    """

    value: Optional[Any] = None
    """
    Optional. Replacement value to use for the response.
    Default: None
    """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .citations import Appearance, ClientCitation, Pattern, SensitivityUsageInfo

__all__ = ["ClientCitation", "Appearance", "SensitivityUsageInfo", "Pattern"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass
class RenderedPromptSection(Generic[T]):
    """
    The result of rendering a section.

    Attributes:
        output (T): The section that was rendered.
        length (int): The number of tokens that were rendered.
        too_long (bool): If true the section was truncated because it exceeded the maxTokens budget.
    """

    output: T
    length: int
    too_long: bool



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import List, cast
from unittest import IsolatedAsyncioTestCase

from botbuilder.core import TurnContext

from teams.ai.augmentations.sequence_augmentation import SequenceAugmentation
from teams.ai.models.chat_completion_action import ChatCompletionAction
from teams.ai.models.prompt_response import PromptResponse
from teams.ai.planners.plan import PredictedDoCommand, PredictedSayCommand
from teams.ai.prompts.message import Message
from teams.ai.prompts.prompt_functions import PromptFunctions
from teams.ai.tokenizers.gpt_tokenizer import GPTTokenizer
from teams.state.turn_state import TurnState


class TestSequenceAugmentation(IsolatedAsyncioTestCase):
    def setUp(self):
        self.test_actions: List[ChatCompletionAction] = [
            ChatCompletionAction(
                name="test1",
                description="test action",
                parameters={
                    "type": "object",
                    "properties": {"foo": {"type": "string"}},
                    "required": ["foo"],
                },
            )
        ]
        self.tokenizer = GPTTokenizer()
        self.sequence_augmentation = SequenceAugmentation(self.test_actions)

    async def test__sequence_aug_create_prompt_section(self):
        state = TurnState()
        section = self.sequence_augmentation.create_prompt_section()
        if section:
            rendered = await section.render_as_messages(
                cast(TurnContext, {}),
                memory=state,
                functions=cast(PromptFunctions, {}),
                tokenizer=self.tokenizer,
                max_tokens=2000,
            )
            self.assertEqual(len(rendered.output), 1)
            self.assertEqual(rendered.output[0].role, "system")
            self.assertNotEqual(
                str(rendered.output[0].content).find(
                    "Use the actions above to create a plan in the following JSON format:"
                ),
                -1,
            )
            self.assertEqual(rendered.too_long, False)

    async def test_validate_badly_formed_plan(self):
        state = TurnState()
        response = await self.sequence_augmentation.validate_response(
            cast(TurnContext, {}),
            state,
            self.tokenizer,
            PromptResponse[str](message=Message(role="assistant", content="")),
            3,
        )
        self.assertEqual(response.valid, False)
        self.assertEqual(
            response.feedback,
            "Return a JSON object that uses the SAY command to say what you're thinking.",
        )

    async def test_validate_empty_action_in_do_command(self):
        state = TurnState()
        response = await self.sequence_augmentation.validate_response(
            cast(TurnContext, {}),
            state,
            self.tokenizer,
            PromptResponse[str](
                message=Message[str](
                    role="assistant",
                    content='{ "type": "plan", '
                    + '"commands": [{ "type": "DO", "action": "", "parameters": {} }]}',
                )
            ),
            3,
        )
        self.assertEqual(response.valid, False)
        self.assertEqual(
            response.feedback,
            'The plan JSON is missing the DO "action" for '
            + "command[0]. Return the name of the action to DO.",
        )

    async def test_validate_invalid_action_in_do_command(self):
        state = TurnState()
        response = await self.sequence_augmentation.validate_response(
            cast(TurnContext, {}),
            state,
            self.tokenizer,
            PromptResponse[str](
                message=Message[str](
                    role="assistant",
                    content='{ "type": "plan", "commands": '
                    + '[{ "type": "DO", "action": "randomAction", "parameters": {}}]}',
                )
            ),
            3,
        )
        self.assertEqual(response.valid, False)
        self.assertEqual(
            response.feedback, 'Unknown action named "randomAction". Specify a valid action name.'
        )

    async def test_validate_empty_response_in_say_command(self):
        state = TurnState()
        response = await self.sequence_augmentation.validate_response(
            cast(TurnContext, {}),
            state,
            self.tokenizer,
            PromptResponse[str](
                message=Message[str](
                    role="assistant",
                    content='{ "type": "plan", ' + '"commands": [{ "type": "SAY"' + "}]}",
                )
            ),
            3,
        )
        self.assertEqual(response.valid, False)
        self.assertEqual(
            response.feedback,
            'The plan JSON is missing the SAY "response" '
            + "for command[0]. Return the response to SAY.",
        )

    async def test_validate_random_command(self):
        state = TurnState()
        response = await self.sequence_augmentation.validate_response(
            cast(TurnContext, {}),
            state,
            self.tokenizer,
            PromptResponse[str](
                message=Message[str](
                    role="assistant", content='{ "type": "plan", "commands": [{ "type": "DOO" }]}'
                )
            ),
            3,
        )
        self.assertEqual(response.valid, False)
        self.assertEqual(
            response.feedback,
            "The JSON returned had errors. Apply these fixes:\n'DOO' is not one of ['DO', 'SAY']",
        )

    async def test_validate_correct_do_and_say_commands(self):
        state = TurnState()
        response = await self.sequence_augmentation.validate_response(
            cast(TurnContext, {}),
            state,
            self.tokenizer,
            PromptResponse[str](
                message=Message[str](
                    role="assistant",
                    content='{"type":"plan","commands":[{"type":"DO",'
                    + '"action":"test1","parameters": { "foo": "bar" }},'
                    + '{"type":"SAY","response": "hello world"}]}',
                )
            ),
            3,
        )
        self.assertEqual(response.feedback, None)
        self.assertEqual(response.valid, True)

    async def test_create_plan_from_response(self):
        state = TurnState()
        # Validate response
        validation = await self.sequence_augmentation.validate_response(
            cast(TurnContext, {}),
            state,
            self.tokenizer,
            PromptResponse[str](
                message=Message[str](
                    role="assistant",
                    content='{"type":"plan","commands":[{"type":"DO",'
                    + '"action":"test1","parameters": { "foo": "bar" }},'
                    + '{"type":"SAY","response": "hello world"}]}',
                )
            ),
            3,
        )

        # Create plan from response
        plan = await self.sequence_augmentation.create_plan_from_response(
            cast(TurnContext, {}),
            state,
            PromptResponse(message=Message(role="assistant", content=validation.value)),
        )
        command0 = cast(PredictedDoCommand, plan.commands[0])
        command1 = cast(PredictedSayCommand, plan.commands[1])

        self.assertEqual(len(plan.commands), 2)
        self.assertEqual(command0.type, "DO")
        self.assertEqual(command0.action, "test1")
        self.assertEqual(command0.parameters, {"foo": "bar"})

        self.assertEqual(command1.type, "SAY")
        assert command1.response is not None
        self.assertEqual(command1.response.role, "assistant")
        self.assertEqual(command1.response.content, "hello world")
